# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import platform
from collections.abc import Generator
from pathlib import Path

import lancedb
import numpy as np
import pandas as pd
import pyarrow as pa
import pytest
from pyarrow.fs import FileSystem, FileType

from geneva import connect, udf
from geneva.packager import DockerUDFPackager

pytestmark = pytest.mark.ray


def test_add_column(tmp_path: Path) -> None:
    db = connect(tmp_path)

    # create a basic table
    tbl = pa.Table.from_pydict({"id": [1, 2, 3, 4, 5, 6]})
    table = db.create_table("table1", tbl)

    # add a basic column
    table.add_columns(
        {"id2": "cast(null as string)"},
    )

    schema = table.schema

    assert len(schema) == 2
    field = schema.field("id2")
    assert field is not None
    assert field.type == pa.string()


def test_add_column_trailng_slash(tmp_path: Path) -> None:
    # make sure we handle trailing slashes the same way
    db = connect(str(tmp_path) + "/")

    # create a basic table
    tbl = pa.Table.from_pydict({"id": [1, 2, 3, 4, 5, 6]})
    table = db.create_table("table1", tbl)

    # add a basic column
    table.add_columns(
        {"id2": "cast(null as string)"},
    )

    schema = table.schema

    assert len(schema) == 2
    field = schema.field("id2")
    assert field is not None
    assert field.type == pa.string()


@pytest.fixture(params=["geneva", "lance"])
def db(tmp_path: Path, request) -> lancedb.DBConnection:
    """Create a temporary database for testing."""
    if request.param == "geneva":
        return connect(tmp_path)
    else:
        return lancedb.connect(tmp_path)


def test_create_table_and_index(db: lancedb.DBConnection) -> None:
    assert len(list(db.table_names())) == 0

    df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5, 6],
            "name": ["Alice", "Bob", "Charlie", "David", "Eve", "Frank"],
        }
    )
    db.create_table("tbl", df)

    assert len(list(db.table_names())) == 1
    assert db.table_names()[0] == "tbl"

    tbl = db.open_table("tbl")
    assert tbl.count_rows() == 6

    df = pd.DataFrame(
        {
            "id": [11, 12, 13, 14, 15, 16],
            "name": ["alice", "bob", "charlie", "david", "eve", "frank"],
        }
    )
    tbl.add(df)
    assert tbl.count_rows() == 12

    tbl.create_scalar_index("id")
    tbl.create_fts_index("name", use_tantivy=False)

    tbl.delete("id % 2 == 0")
    assert tbl.version == 5
    assert tbl.to_arrow().combine_chunks() == pa.Table.from_pydict(
        {
            "id": [1, 3, 5, 11, 13, 15],
            "name": ["Alice", "Charlie", "Eve", "alice", "charlie", "eve"],
        }
    )
    assert tbl.count_rows() == 6

    fts_results = tbl.search("charlie", query_type="fts").to_list()
    assert len(fts_results) == 2  # expect 'charlie' and 'Charlie'


def test_create_vector_idx(db: lancedb.DBConnection) -> None:
    dim = 128

    def producer() -> Generator[pa.Table, None, None]:
        rng = np.random.default_rng()
        for i in range(100):
            ids = pa.array([i * 20 + j for j in range(20)])
            values = pa.array(rng.random(20 * dim).astype(np.float32))
            fsl = pa.FixedSizeListArray.from_arrays(values, dim)  # type: ignore
            yield pa.Table.from_arrays([ids, fsl], ["id", "vector"])

    tbl = db.create_table("table", producer())
    assert tbl.count_rows() == 2000
    tbl.create_index(num_sub_vectors=8)

    indices = tbl.list_indices()
    assert indices[0].index_type == "IvfPq"

    # do a vector search
    rng = np.random.default_rng()
    vec = rng.random(dim)

    # this does not throw an exception
    vec_results = tbl.search(vec).to_list()

    assert len(vec_results) == 10


def test_add_invalid_computed_column(tmp_path: Path) -> None:
    db = connect(tmp_path)

    # create a basic table
    tbl = pa.Table.from_pydict({"id": [1, 2, 3, 4, 5, 6]})
    table = db.create_table("table1", tbl)

    @udf(data_type=pa.int64())
    def double_id(id2: int):  # noqa A002
        return id2 * 2

    # Validation catches missing input column 'id2' before circular dependency check
    with pytest.raises(
        ValueError,
        match=r"expects input columns \['id2'\].*not found in table schema",
    ):
        table.add_columns(
            {"id2": double_id},  # implicit udf arg name mapping
        )


def test_add_circular_dependency_column(tmp_path: Path) -> None:
    """Test that circular dependencies are detected."""
    db = connect(tmp_path)

    # create a basic table
    tbl = pa.Table.from_pydict({"id": [1, 2, 3, 4, 5, 6]})
    table = db.create_table("table1", tbl)

    # First add a column 'id2'
    table.add_columns({"id2": "cast(null as bigint)"})

    # Now try to create a UDF that depends on itself
    @udf(data_type=pa.int64())
    def self_referencing(id2: int) -> int:  # noqa A002
        return id2 * 2

    # This should fail with circular dependency error, not missing column
    with pytest.raises(
        ValueError, match=r"UDF output column id2 is not allowed to be in input"
    ):
        table.add_columns(
            {"id2": self_referencing},  # Column exists, but creates circular dependency
        )


def test_add_computed_column(tmp_path: Path) -> None:
    packager = DockerUDFPackager(
        # use prebuilt image tag so we don't have to build the image
        prebuilt_docker_img="test-image:latest"
    )
    db = connect(tmp_path, packager=packager)

    # create a basic table
    tbl = pa.Table.from_pydict({"id": [1, 2, 3, 4, 5, 6]})
    table = db.create_table("table1", tbl)

    @udf(data_type=pa.int64())
    def double_id(id: int):  # noqa A002
        return id * 2

    table.add_columns(
        {"id2": double_id},  # implicit udf arg name mapping
    )

    schema = table.schema

    assert len(schema) == 2
    field = schema.field("id2")
    assert field is not None
    assert field.type == pa.int64()

    assert len(field.metadata) == 8
    assert field.metadata[b"virtual_column"] == b"true"
    assert field.metadata[b"virtual_column.udf_backend"] == b"DockerUDFSpecV1"
    assert field.metadata[b"virtual_column.udf_name"] == b"double_id"
    assert field.metadata[b"virtual_column.udf_inputs"] == b'["id"]'
    assert field.metadata[
        b"virtual_column.platform.system"
    ] == platform.system().encode("utf-8")
    assert field.metadata[b"virtual_column.platform.arch"] == platform.machine().encode(
        "utf-8"
    )
    assert field.metadata[
        b"virtual_column.platform.python_version"
    ] == platform.python_version().encode("utf-8")

    # check that the UDF was actually uploaded
    fs, root_path = FileSystem.from_uri(f"{str(tmp_path)}/table1.lance")
    file_info = fs.get_file_info(
        f"{root_path}/{field.metadata[b'virtual_column.udf'].decode('utf-8')}"
    )
    assert file_info.type is not FileType.NotFound

    try:
        import ray  # noqa: F401

        # before materializing, the computed column should have nulls
        expected = pa.Table.from_pydict(
            {
                "id": [1, 2, 3, 4, 5, 6],
                "id2": pa.array([None] * 6, pa.int64()),
            }
        )
        assert table.to_arrow().equals(expected)

        # backfill and check values of computed column
        table.backfill("id2")
        assert table.to_pandas().equals(
            pd.DataFrame(
                {
                    "id": [1, 2, 3, 4, 5, 6],
                    "id2": [2, 4, 6, 8, 10, 12],
                }
            )
        )
    except ImportError:
        "Ray not installed, skipping value checks part of test"


def test_alter_computed_column(tmp_path: Path) -> None:
    packager = DockerUDFPackager(
        # use prebuilt image tag so we don't have to build the image
        prebuilt_docker_img="test-image:latest"
    )
    db = connect(tmp_path, packager=packager)

    # create a basic table
    tbl = pa.Table.from_pydict({"id": [1, 2, 3, 4, 5, 6]})
    table = db.create_table("table1", tbl)

    @udf(data_type=pa.int64())
    def double_id(id: int):  # noqa A002
        return id * 2

    table.add_columns(
        {"id2": double_id},  # implicit udf arg name mapping
    )

    schema = table.schema
    field = schema.field("id2")
    assert field.metadata[b"virtual_column.udf_name"] == b"double_id"

    try:
        import kubernetes  # noqa: F401
        import ray  # noqa: F401

        # before materializing, the computed column should have nulls
        expected = pa.Table.from_pydict(
            {
                "id": [1, 2, 3, 4, 5, 6],
                "id2": pa.array([None] * 6, pa.int64()),
            }
        )
        assert table.to_arrow().equals(expected)

        # backfill and check values of computed column
        table.backfill("id2")
        assert table.to_pandas().equals(
            pd.DataFrame(
                {
                    "id": [1, 2, 3, 4, 5, 6],
                    "id2": [2, 4, 6, 8, 10, 12],
                }
            )
        )
    except ImportError:
        "Ray not installed, skipping value checks part of test"

    # now check that we can replace the UDF with a new version:
    @udf(data_type=pa.int64())
    def triple_id(id: int):  # noqa A002
        return id * 3

    # "virtual_column" is deprecated, but still works for backwards compatibility
    table.alter_columns(
        {
            "path": "id2",
            "virtual_column": triple_id,
        }
    )

    table.alter_columns(
        {
            "path": "id2",
            "udf": triple_id,
        }
    )

    schema = table.schema
    field = schema.field("id2")
    assert field.metadata[b"virtual_column"] == b"true"
    assert field.metadata[b"virtual_column.udf_name"] == b"triple_id"
    assert field.metadata[b"virtual_column.udf_backend"] == b"DockerUDFSpecV1"

    # check that the UDF was actually uploaded
    fs, root_path = FileSystem.from_uri(f"{str(tmp_path)}/table1.lance")
    file_info = fs.get_file_info(
        f"{root_path}/{field.metadata[b'virtual_column.udf'].decode('utf-8')}"
    )
    assert file_info.type is not FileType.NotFound

    try:
        import kubernetes  # noqa: F401
        import ray  # noqa: F401

        # After the alter but before materializing, the computed column should
        # not have the old UDF's values.  It should have nulls.
        expected = pa.Table.from_pydict(
            {
                "id": [1, 2, 3, 4, 5, 6],
                "id2": [2, 4, 6, 8, 10, 12],
                #  TODO This should be "id2": pa.array([None] * 6, pa.int64()),
            }
        )
        assert table.to_arrow().equals(expected)

        # backfill and check values of computed column
        table.backfill("id2")
        assert table.to_pandas().equals(
            pd.DataFrame(
                {
                    "id": [1, 2, 3, 4, 5, 6],
                    "id2": [3, 6, 9, 12, 15, 18],
                }
            )
        )
    except ImportError:
        "Ray not installed, skipping value checks part of test"


def test_alter_computed_column_diff_col_name(tmp_path: Path) -> None:
    packager = DockerUDFPackager(
        # use prebuilt image tag so we don't have to build the image
        prebuilt_docker_img="test-image:latest"
    )
    db = connect(tmp_path, packager=packager)

    # create a basic table
    tbl = pa.Table.from_pydict({"seq": [1, 2, 3, 4, 5, 6]})
    table = db.create_table("table1", tbl)

    @udf(data_type=pa.int64())
    def double_id(id: int):  # noqa A002
        return id * 2

    table.add_columns(
        {"id2": (double_id, ["seq"])},  # explicit udf arg name mapping
    )

    schema = table.schema
    field = schema.field("id2")
    assert field.metadata[b"virtual_column.udf_name"] == b"double_id"

    # before materializing, the computed column should have nulls
    expected = pa.Table.from_pydict(
        {
            "seq": [1, 2, 3, 4, 5, 6],
            "id2": pa.array([None] * 6, pa.int64()),
        }
    )
    assert table.to_arrow().equals(expected)

    # backfill and check values of computed column
    table.backfill("id2")
    assert table.to_pandas().equals(
        pd.DataFrame(
            {
                "seq": [1, 2, 3, 4, 5, 6],
                "id2": [2, 4, 6, 8, 10, 12],
            }
        )
    )

    # now check that we can replace the UDF with a new version:
    @udf(data_type=pa.int64())
    def triple_id(id: int):  # noqa A002
        return id * 3

    table.alter_columns(
        *[
            {
                "path": "id2",
                "virtual_column": triple_id,
                "input_columns": ["seq"],
            }
        ]
    )

    schema = table.schema
    field = schema.field("id2")
    assert field.metadata[b"virtual_column"] == b"true"
    assert field.metadata[b"virtual_column.udf_name"] == b"triple_id"
    assert field.metadata[b"virtual_column.udf_backend"] == b"DockerUDFSpecV1"

    # check that the UDF was actually uploaded
    fs, root_path = FileSystem.from_uri(f"{str(tmp_path)}/table1.lance")
    file_info = fs.get_file_info(
        f"{root_path}/{field.metadata[b'virtual_column.udf'].decode('utf-8')}"
    )
    assert file_info.type is not FileType.NotFound

    # After the alter but before materializing, the computed column should
    # not have the old UDF's values.  It should have nulls.
    expected = pa.Table.from_pydict(
        {
            "seq": [1, 2, 3, 4, 5, 6],
            "id2": [2, 4, 6, 8, 10, 12],
            #  TODO This should be "id2": pa.array([None] * 6, pa.int64()),
        }
    )
    assert table.to_arrow().equals(expected)

    # backfill and check values of computed column
    table.backfill("id2")
    assert table.to_pandas().equals(
        pd.DataFrame(
            {
                "seq": [1, 2, 3, 4, 5, 6],
                "id2": [3, 6, 9, 12, 15, 18],
            }
        )
    )



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import logging
import time
import uuid

import pyarrow as pa
import ray

import geneva
from geneva.cluster import K8sConfigMethod
from geneva.cluster.builder import GenevaClusterBuilder
from geneva.manifest.builder import GenevaManifestBuilder
from geneva.runners.ray.raycluster import ClusterStatus, RayCluster
from geneva.utils import dt_now_utc

_LOG = logging.getLogger(__name__)


# use a random version to force checkpoint to invalidate
@geneva.udf(num_cpus=1, version=uuid.uuid4().hex)
def plus_one(a: int) -> int:
    return a + 1


SIZE = 1024


def test_get_imported(
    geneva_test_bucket: str,
    standard_cluster: RayCluster,
) -> None:
    from geneva.runners.ray.pipeline import get_imported

    geneva.connect(geneva_test_bucket)
    with standard_cluster:
        pkgs = ray.get(get_imported.remote())
        for pkg, ver in sorted(pkgs.items()):
            _LOG.info(f"{pkg}=={ver}")


def test_ray_add_column_pipeline(
    geneva_test_bucket: str,
    standard_cluster: RayCluster,
) -> None:
    conn = geneva.connect(geneva_test_bucket)
    table_name = uuid.uuid4().hex
    table = conn.create_table(
        table_name,
        pa.Table.from_pydict({"a": pa.array(range(SIZE))}),
    )
    with standard_cluster:
        table.add_columns(
            {"b": plus_one},  # type: ignore[arg-type]
            batch_size=32,
            concurrency=2,
            intra_applier_concurrency=2,
        )
        table.backfill("b")

    assert table.to_arrow() == pa.Table.from_pydict(
        {"a": pa.array(range(SIZE)), "b": pa.array(range(1, SIZE + 1))}
    )
    conn.drop_table(table_name)


def test_ray_add_column_pipeline_backfill_async(
    geneva_test_bucket: str,
    standard_cluster: RayCluster,
) -> None:
    conn = geneva.connect(geneva_test_bucket)
    table_name = uuid.uuid4().hex
    table = conn.create_table(
        table_name,
        pa.Table.from_pydict({"a": pa.array(range(SIZE))}),
    )
    with standard_cluster:
        table.add_columns(
            {"b": plus_one},  # type: ignore[arg-type]
            batch_size=32,
            concurrency=8,
            intra_applier_concurrency=8,
        )
        fut = table.backfill_async("b")
        while not fut.done():
            time.sleep(1)
        table.checkout_latest()

        time.sleep(10)  # todo: why is this needed?
        _LOG.info("FUT pbars: %s", fut._pbars)  # type: ignore[attr-defined]
        # there should be 4 pbars - geneva, checkpointed, ready to commit and committed
        assert len(fut._pbars) == 4  # type: ignore[attr-defined]

        cs = ClusterStatus()
        cs.get_status()
        assert cs.pbar_k8s is not None
        assert cs.pbar_kuberay is not None

    assert table.to_arrow() == pa.Table.from_pydict(
        {"a": pa.array(range(SIZE)), "b": pa.array(range(1, SIZE + 1))}
    )
    conn.drop_table(table_name)


def test_ray_add_column_pipeline_cpu_only_pool(
    geneva_test_bucket: str,
    standard_cluster: RayCluster,
) -> None:
    conn = geneva.connect(geneva_test_bucket)
    table_name = uuid.uuid4().hex
    table = conn.create_table(
        table_name,
        pa.Table.from_pydict({"a": pa.array(range(SIZE))}),
    )
    with standard_cluster:
        table.add_columns(
            {"b": plus_one},  # type: ignore[arg-type]
            batch_size=32,
            concurrency=4,
            use_cpu_only_pool=True,
        )
        table.backfill("b")

    assert table.to_arrow() == pa.Table.from_pydict(
        {"a": pa.array(range(SIZE)), "b": pa.array(range(1, SIZE + 1))}
    )
    conn.drop_table(table_name)


def test_backfill_multiple_fragments_with_context(
    geneva_test_bucket: str,
    slug: str | None,
    geneva_k8s_service_account: str,
    k8s_namespace: str,
    head_node_selector: dict,
    worker_node_selector: dict,
    region: str,
    k8s_config_method: K8sConfigMethod,
    manifest: str | None,
) -> None:
    adds = 10
    rows_per_add = 10

    db = geneva.connect(geneva_test_bucket)
    table_name = uuid.uuid4().hex
    data = pa.Table.from_pydict({"a": pa.array(range(rows_per_add))})
    table = db.create_table(table_name, data)
    for _ in range(adds):
        # split adds to create many fragments
        data = pa.Table.from_pydict({"a": pa.array(range(rows_per_add))})
        table.add(data)

    cluster_name = "add-column-context"
    cluster_name += f"-{dt_now_utc().strftime('%Y-%m-%d-%H-%M')}-{slug}"
    manifest_name = f"test-manifest-{slug}"

    success = False
    try:
        cluster = (
            GenevaClusterBuilder()
            .name(cluster_name)
            .namespace(k8s_namespace)
            .config_method(k8s_config_method)
            .aws_config(region=region, role_name="geneva-client-role")
            .head_group(
                service_account=geneva_k8s_service_account,
                node_selector=head_node_selector,
                # cpus=2, memory="4Gi", num_gpus=0 use defaults
            )
            .add_cpu_worker_group(
                service_account=geneva_k8s_service_account,
                node_selector=worker_node_selector,
                # cpus=4, memory="8Gi" use defaults
            )
            .add_cpu_worker_group(
                service_account=geneva_k8s_service_account,
                node_selector=worker_node_selector,
                # cpus=4, memory="8Gi" use defaults
            )
            .build()
        )

        db.define_cluster(cluster_name, cluster)

        if manifest:
            # saved manifest provided via --manifest arg
            manifest_name = manifest
        else:
            db.define_manifest(
                manifest_name,
                GenevaManifestBuilder.create(manifest_name)
                .delete_local_zips(True)
                .build(),
            )

        with db.context(
            cluster=cluster_name, manifest=manifest_name, log_to_driver=False
        ):
            table.add_columns(
                {"b": plus_one},  # type: ignore[arg-type]
            )
            table.backfill("b")

            jobs = db._history.list_jobs(table_name, None)
            _LOG.info(f"{jobs=}")
            assert len(jobs) == 1, "expected a job record"
            assert jobs[0].status == "DONE"
            assert jobs[0].table_name == table_name
            assert jobs[0].job_type == "BACKFILL"
            assert jobs[0].metrics, "expected metrics"
            assert jobs[0].events, "expected events"

            success = True

        assert success, "pipeline failed"

    finally:
        db.delete_cluster(cluster_name)
        db.drop_table(table_name)
        if not manifest:
            db.delete_manifest(manifest_name)



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""
Manifest factory for video embeddings UDF.

This module creates GenevaManifest objects by reading dependencies
from pyproject.toml, providing a single source of truth for UDF dependencies.
"""

from pathlib import Path

from geneva.manifest.mgr import GenevaManifest


def create_manifest(name: str | None = None) -> GenevaManifest:
    """
    Create a GenevaManifest for video embeddings UDF.

    Reads dependencies from pyproject.toml and creates a manifest
    that can be uploaded to Ray workers.

    NOTE: This UDF requires torch, torchvision, torchcodec, and transformers
    with FFmpeg system libraries. These MUST be provided by the cluster's
    conda environment (ray_init_kwargs), not via pip, because:
    - torchcodec needs FFmpeg shared libraries (libavutil.so, etc.)
    - Pip installs into isolated virtualenv without conda's system libs

    The manifest only includes minimal dependencies not in the cluster env.

    Args:
        name: Optional manifest name. If not provided, uses project name from pyproject.toml

    Returns:
        GenevaManifest configured with minimal dependencies
    """
    # Python 3.10 compatibility - use tomli backport if tomllib not available
    try:
        import tomllib
    except ModuleNotFoundError:
        import tomli as tomllib  # type: ignore[no-redef]

    udf_dir = Path(__file__).parent
    pyproject_path = udf_dir / "pyproject.toml"

    with pyproject_path.open("rb") as f:
        pyproject = tomllib.load(f)

    manifest_name = name or pyproject["project"]["name"]

    # Return empty pip dependencies - all deps provided by cluster's conda environment
    # Ray doesn't allow mixing 'pip' and 'conda' fields in runtime_env, so when
    # using a cluster with conda (ray_init_kwargs), the manifest must have NO pip deps.
    return GenevaManifest(
        name=manifest_name,
        pip=[],  # Empty - all deps in cluster's conda environment
        py_modules=[],  # Empty - UDF modules are in zips, extracted via GENEVA_ZIPS
        skip_site_packages=True,  # Do not include site-packages
        delete_local_zips=True,
    )



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import base64
import os


def test_geneva_zips_env_var() -> None:
    """
    Test that the GENEVA_ZIPS environment variable is set correctly.
    """
    os.environ["GENEVA_ZIPS"] = base64.b64encode(b'{"zips": [[]]}').decode("utf-8")
    # this loads data if GENEVA_ZIPS is set and could blow up
    import geneva  # noqa: F401



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors
import tempfile
import uuid
from datetime import timezone
from pathlib import Path
from typing import Any

import pytest
import ray
from lancedb import connect_async

from geneva.db import connect
from geneva.jobs.jobs import (
    GENEVA_JOBS_TABLE_NAME,
    JobMetric,
    JobRecord,
    JobStateManager,
)
from geneva.runners.ray.jobtracker import JobTracker
from geneva.table import TableReference

pytestmark = pytest.mark.ray


@pytest.fixture(autouse=True)
def ray_cluster() -> None:
    ray.shutdown()
    ray.init(
        log_to_driver=True,
        logging_config=ray.LoggingConfig(
            encoding="TEXT", log_level="DEBUG", additional_log_standard_attrs=["name"]
        ),
    )
    yield
    ray.shutdown()


@pytest.fixture
async def temp_db_path() -> Any:
    """Create a temporary database directory."""
    with tempfile.TemporaryDirectory() as temp_dir:
        yield Path(temp_dir)


@pytest.fixture
async def db_connection(temp_db_path) -> Any:
    """Create a Geneva database connection."""
    db = connect(temp_db_path)
    yield db
    db.close()


@pytest.fixture
async def jobs_table(db_connection) -> Any:
    """Create the geneva_jobs table with proper schema."""
    # Create the jobs table using JobStateManager to ensure proper schema
    job_manager = JobStateManager(db_connection)
    return job_manager.get_table()


@pytest.fixture
async def table_reference(temp_db_path) -> TableReference:
    """Create a TableReference for testing."""
    return TableReference(
        db_uri=str(temp_db_path), table_name="test_table", version=None
    )


@pytest.fixture
async def async_db_connection(temp_db_path) -> Any:
    """Create an async LanceDB connection."""
    async_conn = await connect_async(str(temp_db_path))
    yield async_conn
    if hasattr(async_conn, "close"):
        close_result = async_conn.close()
        if hasattr(close_result, "__await__"):
            await close_result


def test_jobtracker_creation(table_reference) -> None:
    """Test JobTracker can be created with required fields."""
    job_id = str(uuid.uuid4())

    # Create JobTracker instance (without Ray remote for testing)
    tracker = JobTracker.remote(job_id, table_reference)

    tracker.get_all.remote()


@pytest.mark.asyncio
async def test_save_metrics_with_mock_db(
    temp_db_path, async_db_connection, db_connection
) -> None:
    """Test _save_metrics with a real async database connection."""
    job_id = str(uuid.uuid4())

    # Create the jobs table
    jsm = JobStateManager(db_connection, GENEVA_JOBS_TABLE_NAME)

    # Add a test job record
    from datetime import datetime

    jobs_table = await async_db_connection.open_table(GENEVA_JOBS_TABLE_NAME)
    time = datetime(2025, 11, 7, 9, 46, 9, 599847, tzinfo=timezone.utc)
    await jobs_table.add(
        [
            {
                "job_id": job_id,
                "table_name": "test_table",
                "column_name": "test_column",
                "status": "RUNNING",
                "metrics": [],
                "launched_at": time,
                "completed_at": None,
                "config": "{}",
                "launched_by": "test",
                "manifest_id": None,
                "manifest_checksum": None,
                "events": [],
                "object_ref": None,
                "job_type": "BACKFILL",
            }
        ]
    )

    # Create JobTracker and manually set up the connection
    table_ref = TableReference(
        db_uri=str(temp_db_path), table_name="test_table", version=None
    )

    tracker = JobTracker.remote(job_id, table_ref)

    # Manually set the database connection for testing
    tracker._db = async_db_connection
    tracker._jobs_table = jobs_table

    # Test _save_metrics
    test_metrics = {
        "task1": {"n": 50, "total": 100, "done": False, "desc": "Test task"}
    }

    await tracker._save_metrics.remote(test_metrics)

    jobs = jsm.list_jobs(table_name="test_table")
    assert jobs == [
        JobRecord(
            table_name="test_table",
            column_name="test_column",
            job_id=job_id,
            job_type="BACKFILL",
            object_ref=None,
            status="RUNNING",
            launched_at=time,
            completed_at=None,
            config="{}",
            launched_by="test",
            manifest_id=None,
            manifest_checksum=None,
            metrics=[
                JobMetric(name="task1", n=50, total=100, done=False, desc="Test task")
            ],
            events=[],
        )
    ]


@pytest.mark.asyncio
async def test_full_workflow_with_db(temp_db_path, db_connection) -> None:
    """Test a complete workflow with real database operations."""
    job_id = str(uuid.uuid4())

    # Set up the database with jobs table
    job_manager = JobStateManager(db_connection)
    job_record = job_manager.launch("test_table", "test_column")

    # Override with our test job_id
    job_manager.get_table().update(
        where=f"job_id = '{job_record.job_id}'", values={"job_id": job_id}
    )

    # Create async connection for JobTracker
    async_conn = await connect_async(str(temp_db_path))
    jobs_table = await async_conn.open_table(GENEVA_JOBS_TABLE_NAME)

    # Create JobTracker and set up connection
    table_ref = TableReference(
        db_uri=str(temp_db_path), table_name="test_table", version=None
    )

    tracker = JobTracker.remote(job_id, table_ref)
    tracker._db = async_conn
    tracker._jobs_table = jobs_table

    # Test complete workflow
    await tracker.set_total.remote("download", 1000)
    await tracker.set_desc.remote("download", "foo")

    # Simulate progress
    for i in range(0, 1001, 100):
        await tracker.set.remote("download", i)
        if i == 1000:
            break

    # mark done to flush to db
    tracker.mark_done.remote("done")

    # Verify final state
    progress = await tracker.get_progress.remote("download")
    assert progress["n"] == 1000
    assert progress["total"] == 1000
    assert progress["done"] is True
    assert progress["desc"] == "foo"

    # Verify metrics updated in database
    jobs = job_manager.get(job_id)
    assert len(jobs) == 1
    stored_metrics = jobs[0].metrics
    assert len(stored_metrics) == 1

    metric_data = stored_metrics[0]
    assert metric_data.name == "download"
    assert metric_data.n == 1000
    assert metric_data.total == 1000
    assert metric_data.done is True


def test_save_with_throttle_logic() -> None:
    """Test the throttling logic directly."""
    current_time = 0.0
    save_calls = []

    def mock_get_time() -> float:
        return current_time

    def mock_save(metrics) -> None:  # noqa: ARG001
        save_calls.append(current_time)

    # Simulate the throttle logic
    min_time_between_updates = 5.0
    last_updated = -float("inf")  # Initialize to allow first save

    def save_with_throttle(force: bool = False) -> None:
        nonlocal last_updated
        if not force and last_updated + min_time_between_updates > current_time:
            return
        last_updated = current_time
        mock_save({})

    # Test: First save at time 0
    current_time = 0.0
    save_with_throttle()
    assert len(save_calls) == 1

    # Test: Save at time 2 (within throttle) - should be blocked
    current_time = 2.0
    save_with_throttle()
    assert len(save_calls) == 1  # Still 1

    # Test: Save at time 6 (beyond throttle) - should save
    current_time = 6.0
    save_with_throttle()
    assert len(save_calls) == 2

    # Test: Force save at time 7 (within throttle) - should save anyway
    current_time = 7.0
    save_with_throttle(force=True)
    assert len(save_calls) == 3


def test_completion_forces_save_logic() -> None:
    """Test that completion forces save."""
    save_calls = []
    current_time = 0.0

    def mock_save(metrics) -> None:  # noqa: ARG001
        save_calls.append(current_time)

    min_time_between_updates = 100.0  # Very long throttle
    last_updated = -float("inf")

    def save_with_throttle(force: bool = False) -> None:
        nonlocal last_updated
        if not force and last_updated + min_time_between_updates > current_time:
            return
        last_updated = current_time
        mock_save({})

    # First save
    current_time = 0.0
    save_with_throttle()
    assert len(save_calls) == 1

    # Try to save within throttle window - should be blocked
    current_time = 1.0
    save_with_throttle(force=False)
    assert len(save_calls) == 1

    # Force save (like when completing) - should save
    save_with_throttle(force=True)
    assert len(save_calls) == 2


def test_set_and_increment_logic() -> None:
    """Test the set/increment logic for determining when to force save."""

    # Test increment that completes
    total = 100
    n = 90
    done = False
    if total and n >= total:
        done = True
    assert done is False

    # After increment
    n += 10
    if total and n >= total:
        done = True
    assert done is True

    # Test set that completes
    total = 100
    n = 0
    done = False
    n = 100
    if total and n >= total:
        done = True
    assert done is True

    # Test increment past total
    total = 100
    n = 0
    n += 150
    done = False
    if total and n >= total:
        done = True
    assert done is True


def test_zero_throttle() -> None:
    """Test that zero throttle allows all saves."""
    save_calls = []
    current_time = 0.0

    def mock_save(metrics) -> None:  # noqa: ARG001
        save_calls.append(current_time)

    min_time_between_updates = 0.0  # No throttling
    last_updated = -float("inf")

    def save_with_throttle(force: bool = False) -> None:
        nonlocal last_updated
        if not force and last_updated + min_time_between_updates > current_time:
            return
        last_updated = current_time
        mock_save({})

    # Multiple rapid saves - all should go through
    for i in range(5):
        current_time = i * 0.01
        save_with_throttle()

    assert len(save_calls) == 5


def test_mark_done_always_forces() -> None:
    """Test that mark_done behavior always forces save."""
    # mark_done calls save_with_throttle(force=True)
    # So it should always save regardless of throttle

    save_calls = []
    current_time = 0.0

    def mock_save(metrics) -> None:  # noqa: ARG001
        save_calls.append(current_time)

    min_time_between_updates = 100.0
    last_updated = -float("inf")

    def save_with_throttle(force: bool = False) -> None:
        nonlocal last_updated
        if not force and last_updated + min_time_between_updates > current_time:
            return
        last_updated = current_time
        mock_save({})

    # Initial save
    current_time = 0.0
    save_with_throttle()
    assert len(save_calls) == 1

    # mark_done would call with force=True even within throttle window
    current_time = 1.0
    save_with_throttle(force=True)
    assert len(save_calls) == 2



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# definition of the read task, which is portion of a fragment

import hashlib
import logging
from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import Any, cast

import attrs
import pyarrow as pa
import pyarrow.compute as pc
from typing_extensions import override

from geneva.db import connect
from geneva.query import ExtractedTransform
from geneva.table import Table, TableReference
from geneva.transformer import BACKFILL_SELECTED, UDF, UDFArgType

_LOG = logging.getLogger(__name__)

DEFAULT_CHECKPOINT_ROWS = 100


class ReadTask(ABC):
    """
    A task to read data that has a defined output location and unique identifier
    """

    @abstractmethod
    def to_batches(
        self, *, batch_size=DEFAULT_CHECKPOINT_ROWS
    ) -> Iterator[pa.RecordBatch]:
        """Return the data to read"""

    @abstractmethod
    def checkpoint_key(self) -> str:
        """Return a unique key for this task"""

    @abstractmethod
    def dest_frag_id(self) -> int:
        """Return the id of the destination fragment"""

    @abstractmethod
    def dest_offset(self) -> int:
        """Return the offset into the destination fragment"""

    @abstractmethod
    def num_rows(self) -> int:
        """Return the number of rows this task will read"""


@attrs.define(order=True)
class ScanTask(ReadTask):
    uri: str
    columns: list[str]
    frag_id: int
    offset: int
    limit: int

    version: int | None = None
    where: str | None = None

    with_row_address: bool = False

    def _get_table(self) -> Table:
        # here's an example: "gs://my-bucket/dir1/dir2/data.lance"
        uri_parts = self.uri.split("/")
        tablename = ".".join(uri_parts[-1].split(".")[:-1])  # "data"
        db_path = "/".join(uri_parts[:-1])  # "gs://my-bucket/dir1/dir2"

        tbl = connect(db_path).open_table(tablename, version=self.version)
        return tbl

    @override
    def to_batches(
        self, *, batch_size=DEFAULT_CHECKPOINT_ROWS
    ) -> Iterator[pa.RecordBatch]:
        _LOG.debug(
            f"Reading {self.uri} with version {self.version} for cols {self.columns}"
            f" offset {self.offset} limit {self.limit} where='{self.where}'"
        )
        tbl = self._get_table()
        query = tbl.search().enable_internal_api()  # type: ignore[attr-defined]

        if self.with_row_address:
            query = query.with_row_address()

        query = query.with_fragments(self.frag_id).offset(self.offset).limit(self.limit)
        query = query.with_where_as_bool_column()

        # works with blobs but not filters
        if self.columns is not None:
            query = query.select(self.columns)
        if self.where is not None:
            query = query.where(self.where)

        # Currently lancedb reports the wrong type for the return value
        # of the to_batches method.  Remove pyright ignore when fixed.
        batches: pa.RecordBatchReader = query.to_batches(batch_size)  # pyright: ignore[reportAssignmentType]

        yield from batches

    @override
    def checkpoint_key(self) -> str:
        hasher = hashlib.md5()
        hasher.update(
            f"{self.uri}:{self.version}:{self.columns}:{self.frag_id}:{self.offset}:{self.limit}:{self.where}".encode(),
        )
        return hasher.hexdigest()

    @override
    def dest_frag_id(self) -> int:
        return self.frag_id

    @override
    def dest_offset(self) -> int:
        return self.offset

    @override
    def num_rows(self) -> int:
        # TODO calculate # rows
        return self.limit


@attrs.define(order=True)
class CopyTask(ReadTask):
    src: TableReference
    dst: TableReference
    columns: list[str]
    frag_id: int
    offset: int
    limit: int

    @override
    def to_batches(
        self, *, batch_size=DEFAULT_CHECKPOINT_ROWS
    ) -> Iterator[pa.RecordBatch]:
        dst_tbl = self.dst.open()

        # Read __source_row_id from the specific destination fragment using Lance API
        # This ensures we read from the correct fragment, not from the entire table
        dst_lance = dst_tbl.to_lance()
        dst_fragment = dst_lance.get_fragment(self.frag_id)

        # Use dataset scanner with fragment restriction for efficient offset/limit
        # This only reads the requested slice, avoiding loading the entire fragment
        scanner = dst_lance.scanner(
            columns=["__source_row_id"],
            offset=self.offset,
            limit=self.limit,
            fragments=[dst_fragment],
        )
        row_ids_batch = scanner.to_table()
        row_ids = cast("list[int]", row_ids_batch["__source_row_id"].to_pylist())

        # TODO: Add streaming take to lance
        src_table_lance = self.src.open().to_lance()
        _LOG.info(
            f"CopyTask: frag_id={self.frag_id}, offset={self.offset}, "
            f"limit={self.limit}, row_ids={row_ids}"
        )
        table = src_table_lance._take_rows(row_ids, columns=self.columns)
        _LOG.info(f"CopyTask: Fetched {table.num_rows} rows from source")

        # Generate row addresses based on ACTUAL number of rows returned
        # (not self.limit, which may be larger than actual rows)
        row_addrs = self._get_row_addrs(table.num_rows)
        table = table.add_column(table.num_columns, "_rowaddr", row_addrs)

        batches = table.to_batches(max_chunksize=batch_size)

        yield from batches

    @override
    def checkpoint_key(self) -> str:
        hasher = hashlib.md5()
        hasher.update(
            f"CopyTask:{self.src.db_uri}:{self.src.table_name}:{self.src.version}:{self.columns}:{self.dst.db_uri}:{self.dst.table_name}:{self.frag_id}:{self.offset}:{self.limit}".encode(),
        )
        return hasher.hexdigest()

    @override
    def dest_frag_id(self) -> int:
        return self.frag_id

    @override
    def dest_offset(self) -> int:
        return self.offset

    def _get_row_addrs(self, num_rows: int) -> pa.Array:
        frag_mod = self.frag_id << 32
        addrs = [frag_mod + x for x in range(self.offset, self.offset + num_rows)]
        return cast("pa.Array", pa.array(addrs, pa.uint64()))

    @override
    def num_rows(self) -> int:
        # TODO calculate # rows
        return self.limit


class MapTask(ABC):
    @abstractmethod
    def checkpoint_key(self) -> str:
        """Return a unique name for the task"""

    @abstractmethod
    def name(self) -> str:
        """Return a name to use for progress strings"""

    @abstractmethod
    def apply(self, batch: pa.RecordBatch) -> pa.RecordBatch:
        """Apply the map function to the input batch, returning the output batch"""

    @abstractmethod
    def output_schema(self) -> pa.Schema:
        """Return the output schema"""

    @abstractmethod
    def is_cuda(self) -> bool:
        """Return true if the task requires CUDA

        .. deprecated::
            Use :meth:`num_gpus` instead to get the actual GPU requirement.
        """

    @abstractmethod
    def num_cpus(self) -> float | None:
        """Return the number of CPUs the task should use (None for default)"""

    @abstractmethod
    def num_gpus(self) -> float | None:
        """Return the number of GPUs the task should use (None for default)"""

    @abstractmethod
    def memory(self) -> int | None:
        """Return the amount of RAM the task should use (None for default)"""

    @abstractmethod
    def batch_size(self) -> int:
        """Return the batch size the task should use"""


@attrs.define(order=True)
class BackfillUDFTask(MapTask):
    udfs: dict[str, UDF] = (
        attrs.field()
    )  # TODO: use attrs to enforce stateful udfs are handled here

    # this is needed to differentiate a filtered task's checkpont keys
    # for backfill jobs
    where: str | None = attrs.field(default=None)

    # If set, this overrides the UDF-declared batch size. Used to respect
    # the backfill(batch_size=...) parameter from the job config.
    override_batch_size: int | None = attrs.field(default=None)

    def __get_udf(self) -> tuple[str, UDF]:
        # TODO: Add support for multiple columns to add_columns operation
        if len(self.udfs) != 1:
            raise NotImplementedError("Add columns does not support multiple UDFs")
        col, udf = next(iter(self.udfs.items()))
        if not isinstance(udf, UDF):
            # stateful udf are Callable classes that need to be instantiated.
            udf = udf()
        return col, udf

    @override
    def name(self) -> str:
        name, _ = self.__get_udf()
        return name

    @override
    def checkpoint_key(self) -> str:
        # 'where' is required in key to distinguish different partial backfills jobs
        # hashing it so that it cannot be used as a directory path attack vector
        _, udf = self.__get_udf()
        if self.where:
            hasher = hashlib.md5()
            hasher.update((self.where or "").encode())
            return f"{udf.checkpoint_key}:where={hasher.hexdigest()}"
        return udf.checkpoint_key

    @override
    def apply(self, batch: pa.RecordBatch | list[dict[str, Any]]) -> pa.RecordBatch:
        udf_col_name, udf = self.__get_udf()

        if isinstance(batch, pa.RecordBatch):
            row_addr = batch["_rowaddr"]
            has_carry_forward_col = udf_col_name in batch.schema.names
            has_backfill_selected = BACKFILL_SELECTED in batch.schema.names
        else:
            # might have blob_columns which needs _rowaddr
            row_addr = pa.array([x["_rowaddr"] for x in batch], type=pa.uint64())
            has_carry_forward_col = bool(batch) and (udf_col_name in batch[0])
            has_backfill_selected = bool(batch) and (BACKFILL_SELECTED in batch[0])

        # drop carry_foward cols from what will be UDF arguments.
        if isinstance(batch, pa.RecordBatch):
            if udf_col_name in batch.schema.names:
                # select all the other columns
                col_val_arrays = [
                    batch[col] for col in batch.schema.names if col != udf_col_name
                ]
                col_names = [col for col in batch.schema.names if col != udf_col_name]
                batch_for_udf = pa.RecordBatch.from_arrays(col_val_arrays, col_names)
            else:
                batch_for_udf = batch
        else:
            # list-of-dicts case: just filter out the key
            batch_for_udf = [
                {k: v for k, v in row.items() if k != udf_col_name} for row in batch
            ]

        # execute the udf
        # Optimization: For RecordBatch and Array UDFs with filtering, only process
        # selected rows
        try:
            if (
                has_backfill_selected
                and isinstance(batch, pa.RecordBatch)
                and udf.arg_type in (UDFArgType.RECORD_BATCH, UDFArgType.ARRAY)
            ):
                # Get the selection mask
                mask = batch[BACKFILL_SELECTED]

                # Will be set to array data or None if no rows processed
                new_arr = None

                # Check if any rows are selected
                if any(mask.to_pylist()):
                    # Filter batch_for_udf to only include selected rows
                    filtered_batch_for_udf = pc.filter(batch_for_udf, mask)  # type: ignore[call-overload,arg-type]
                    _LOG.debug(
                        f"{udf.arg_type.name} UDF optimization: processing "
                        f"{filtered_batch_for_udf.num_rows} rows instead of "  # type: ignore[attr-defined]
                        f"{batch_for_udf.num_rows}"  # type: ignore[attr-defined]
                    )

                    # Execute UDF on filtered batch
                    filtered_new_arr = udf(filtered_batch_for_udf, use_applier=True)

                    # Expand results back to full batch size.
                    # Get indices of rows that passed the filter
                    indices_array = pa.array(range(batch.num_rows))
                    selected_indices = pc.filter(indices_array, mask)

                    if len(filtered_new_arr) > 0:
                        # Build complete list with values at correct positions
                        result_pylist = [None] * batch.num_rows
                        for i, filtered_val in enumerate(filtered_new_arr):
                            original_idx = selected_indices[i].as_py()  # type: ignore[attr-defined]
                            result_pylist[original_idx] = filtered_val.as_py()

                        # Create array using pa.table() to ensure proper buffer
                        # structure for variable-width types (strings, binary,
                        # lists). pa.table() guarantees correct buffer allocation.
                        # See writer.py:_make_filler_batch() for details.
                        temp_table = pa.table(
                            {"_temp": result_pylist},
                            schema=pa.schema([("_temp", udf.data_type)]),
                        )  # type: ignore[arg-type,list-item]
                        new_arr = temp_table.column("_temp").combine_chunks()

                # If no rows were processed, return array of nulls
                if new_arr is None:
                    # Use pa.nulls() to ensure proper buffer structure for
                    # variable-width types (strings, binary, lists).
                    # See writer.py:_make_filler_batch() for details.
                    new_arr = pa.nulls(batch.num_rows, type=udf.data_type)  # type: ignore[arg-type]
            else:
                # Original behavior for non-RecordBatch UDFs or when no filtering
                new_arr = udf(batch_for_udf, use_applier=True)
        except KeyError as e:
            # Column not found in batch
            if isinstance(batch_for_udf, pa.RecordBatch):
                available_cols = batch_for_udf.schema.names
            elif batch_for_udf and len(batch_for_udf) > 0:  # list[dict] with elements
                available_cols = list(batch_for_udf[0].keys())
            else:  # empty list
                available_cols = []
            raise KeyError(
                f"UDF '{udf.name}' failed: column {e} not found in batch. "
                f"Available columns: {available_cols}. "
                f"UDF expects input_columns: {udf.input_columns}. "
                f"This typically means the UDF's input_columns don't match "
                f"the table schema."
            ) from e
        except (pa.ArrowInvalid, pa.ArrowTypeError) as e:
            # Type mismatch or serialization error
            batch_schema = (
                batch_for_udf.schema
                if isinstance(batch_for_udf, pa.RecordBatch)
                else None
            )
            raise TypeError(
                f"UDF '{udf.name}' failed with type error: {e}. "
                f"Input batch schema: {batch_schema}. "
                f"UDF expects input_columns: {udf.input_columns}. "
                f"UDF output type: {udf.data_type}. "
                f"This often indicates a type mismatch (e.g., float32 vs float64) "
                f"between the table schema and UDF expectations."
            ) from e

        # now finalize the result.
        if not has_carry_forward_col or not has_backfill_selected:
            schema = pa.schema(
                [
                    pa.field(udf_col_name, udf.data_type, metadata=udf.field_metadata),
                    pa.field("_rowaddr", pa.uint64()),
                ]
            )

            # no carry forward col? return the new
            return pa.record_batch([new_arr, row_addr], schema=schema)

        # handle carry forward of old values
        if isinstance(batch, pa.RecordBatch):
            orig_arr = batch[udf_col_name]
            mask = batch[BACKFILL_SELECTED]
        else:
            # extract py values and wrap into pa array
            orig_vals = [x.get(udf_col_name) for x in batch]
            orig_arr = pa.array(orig_vals, type=new_arr.type)
            mask_vals = [x.get(BACKFILL_SELECTED) for x in batch]
            mask = pa.array(mask_vals, type=pa.bool_())

        merged = pc.if_else(mask, new_arr, orig_arr)
        schema = pa.schema(
            [
                pa.field(udf_col_name, udf.data_type, metadata=udf.field_metadata),
                pa.field("_rowaddr", pa.uint64()),
            ]
        )
        return pa.record_batch([merged, row_addr], schema=schema)

    @override
    def output_schema(self) -> pa.Schema:
        name, udf = self.__get_udf()
        return pa.schema(
            [pa.field(name, udf.data_type), pa.field("_rowaddr", pa.uint64())]
        )

    @override
    def is_cuda(self) -> bool:
        """Deprecated: Use num_gpus() instead."""
        _, udf = self.__get_udf()
        return bool(udf.num_gpus and udf.num_gpus > 0)

    @override
    def num_cpus(self) -> float | None:
        _, udf = self.__get_udf()
        return udf.num_cpus

    @override
    def num_gpus(self) -> float | None:
        _, udf = self.__get_udf()
        return udf.num_gpus

    @override
    def memory(self) -> int | None:
        _, udf = self.__get_udf()
        return udf.memory

    @override
    def batch_size(self) -> int:
        if self.override_batch_size is not None:
            return self.override_batch_size
        _, udf = self.__get_udf()
        return udf.batch_size or DEFAULT_CHECKPOINT_ROWS


@attrs.define(order=True)
class CopyTableTask(MapTask):
    column_udfs: list[ExtractedTransform] = attrs.field()
    view_name: str = attrs.field()
    schema: pa.Schema = attrs.field()

    @override
    def name(self) -> str:
        return self.view_name

    @override
    def checkpoint_key(self) -> str:
        return self.view_name

    @override
    def apply(self, batch: pa.RecordBatch) -> pa.RecordBatch:
        for transform in self.column_udfs:
            new_arr = transform.udf(batch)
            batch = batch.add_column(
                transform.output_index, transform.output_name, new_arr
            )
        return batch

    @override
    def output_schema(self) -> pa.Schema:
        return self.schema

    @override
    def is_cuda(self) -> bool:
        """Deprecated: Use num_gpus() instead."""
        return any(
            column_udf.udf.num_gpus and column_udf.udf.num_gpus > 0
            for column_udf in self.column_udfs
        )

    @override
    def num_cpus(self) -> float | None:
        return max(
            (
                column_udf.udf.num_cpus
                for column_udf in self.column_udfs
                if column_udf.udf.num_cpus is not None
            ),
            default=None,
        )

    @override
    def num_gpus(self) -> float | None:
        return max(
            (
                column_udf.udf.num_gpus
                for column_udf in self.column_udfs
                if column_udf.udf.num_gpus is not None
            ),
            default=None,
        )

    @override
    def memory(self) -> int | None:
        return max(
            (
                column_udf.udf.memory
                for column_udf in self.column_udfs
                if column_udf.udf.memory is not None
            ),
            default=None,
        )

    @override
    def batch_size(self) -> int:
        if not self.column_udfs:
            return DEFAULT_CHECKPOINT_ROWS
        return min(
            column_udf.udf.batch_size or DEFAULT_CHECKPOINT_ROWS
            for column_udf in self.column_udfs
        )



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""Error storage and retry configuration for UDF execution"""

import enum
import logging
import traceback
import uuid
from collections.abc import Callable
from datetime import datetime
from typing import Any, Optional

import attrs
import pyarrow as pa
from tenacity import (
    RetryCallState,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from tenacity.retry import retry_base
from tenacity.stop import stop_base
from tenacity.wait import wait_base

from geneva.state.manager import BaseManager
from geneva.utils import dt_now_utc, escape_sql_string, retry_lance

_LOG = logging.getLogger(__name__)

GENEVA_ERRORS_TABLE_NAME = "geneva_errors"


class FaultIsolation(enum.Enum):
    """Strategy for isolating UDF failures"""

    FAIL_BATCH = "fail_batch"  # Fail entire batch on any error (default)
    SKIP_ROWS = "skip_rows"  # Skip individual failing rows (scalar UDFs only)


@attrs.define
class UDFRetryConfig:
    """Retry configuration for UDF execution using tenacity semantics"""

    # Tenacity retry condition - which exceptions to retry
    retry: retry_base = attrs.field(
        factory=lambda: retry_if_exception_type(())  # No retries by default
    )

    # Stop condition - when to give up
    stop: stop_base = attrs.field(factory=lambda: stop_after_attempt(1))

    # Wait strategy - how long to wait between retries
    wait: wait_base = attrs.field(
        factory=lambda: wait_exponential(multiplier=1, min=1, max=60)
    )

    # Optional callbacks
    before_sleep: Callable[[RetryCallState], None] | None = attrs.field(default=None)
    after_attempt: Callable[[RetryCallState], None] | None = attrs.field(default=None)

    # Whether to reraise exception after retries exhausted
    reraise: bool = attrs.field(default=True)

    @classmethod
    def no_retry(cls) -> "UDFRetryConfig":
        """No retries - fail immediately (default behavior)"""
        return cls()

    @classmethod
    def retry_transient(cls, max_attempts: int = 3) -> "UDFRetryConfig":
        """Retry common transient errors (network, timeouts)

        Parameters
        ----------
        max_attempts : int
            Maximum number of attempts including the initial try
        """
        return cls(
            retry=retry_if_exception_type((OSError, TimeoutError, ConnectionError)),
            stop=stop_after_attempt(max_attempts),
            wait=wait_exponential(multiplier=1, min=1, max=60),
        )


@attrs.define
class ErrorHandlingConfig:
    """Configuration for UDF error handling behavior"""

    # Retry policy using tenacity
    retry_config: UDFRetryConfig = attrs.field(factory=UDFRetryConfig.no_retry)

    # How to isolate failures
    fault_isolation: FaultIsolation = attrs.field(default=FaultIsolation.FAIL_BATCH)

    # Whether to log errors to the error table
    log_errors: bool = attrs.field(default=True)

    # Whether to log all retry attempts (not just final failures)
    log_retry_attempts: bool = attrs.field(default=False)

    def validate_compatibility(self, map_task) -> None:
        """Validate that this error config is compatible with the given task

        Args:
            map_task: The MapTask to validate against

        Raises:
            ValueError: If SKIP_ROWS is used with RecordBatch UDF
        """
        from geneva.apply.task import BackfillUDFTask
        from geneva.transformer import UDFArgType

        if self.fault_isolation != FaultIsolation.SKIP_ROWS:
            return

        # SKIP_ROWS only works with scalar/array UDFs, not RecordBatch UDFs
        if isinstance(map_task, BackfillUDFTask):
            _, udf = next(iter(map_task.udfs.items()))
            if hasattr(udf, "arg_type") and udf.arg_type == UDFArgType.RECORD_BATCH:
                raise ValueError(
                    "SKIP_ROWS fault isolation cannot be used with "
                    "RecordBatch UDFs. RecordBatch UDFs process entire "
                    "batches and cannot skip individual rows. "
                    "Use FAIL_BATCH instead."
                )


@attrs.define(kw_only=True)
class ErrorRecord:
    """UDF execution error record, stored in geneva_errors table"""

    # Unique error ID
    error_id: str = attrs.field(factory=lambda: str(uuid.uuid4()))

    # Error details
    error_type: str = attrs.field()  # Exception.__class__.__name__
    error_message: str = attrs.field()
    error_trace: str = attrs.field()  # Full traceback

    # Job/Table context
    job_id: str = attrs.field()
    table_uri: str = attrs.field()  # Full URI to the table
    table_name: str = attrs.field()
    table_version: Optional[int] = attrs.field(default=None)  # Read version
    column_name: str = attrs.field()

    # UDF context
    udf_name: str = attrs.field()
    udf_version: str = attrs.field()

    # Execution context (Ray/distributed)
    actor_id: Optional[str] = attrs.field(default=None)
    fragment_id: Optional[int] = attrs.field(default=None)
    batch_index: int = attrs.field()  # Sequence number within fragment

    # Row-level granularity (for scalar UDFs)
    row_address: Optional[int] = attrs.field(default=None)

    # Retry context
    attempt: int = attrs.field(default=1)
    max_attempts: int = attrs.field(default=1)

    # Timestamp
    timestamp: datetime = attrs.field(
        factory=dt_now_utc, metadata={"pa_type": pa.timestamp("us", tz="UTC")}
    )


class ErrorStore(BaseManager):
    """Store and query error records in a Lance table"""

    def get_model(self) -> Any:
        return ErrorRecord(
            error_type="InitError",
            error_message="init",
            error_trace="",
            job_id="init",
            table_uri="init",
            table_name="init",
            column_name="init",
            udf_name="init",
            udf_version="init",
            batch_index=0,
            table_version=0,
        )

    def get_table_name(self) -> str:
        return GENEVA_ERRORS_TABLE_NAME

    @retry_lance
    def log_error(self, error: ErrorRecord) -> None:
        """Log an error record to the error table

        Parameters
        ----------
        error : ErrorRecord
            The error record to log
        """
        self.get_table().add([attrs.asdict(error)])
        _LOG.info(
            f"Logged error {error.error_id}: {error.error_type} in "
            f"{error.table_name}.{error.column_name} (attempt {error.attempt})"
        )

    @retry_lance
    def log_errors(self, errors: list[ErrorRecord]) -> None:
        """Log multiple error records to the error table in a single operation

        Parameters
        ----------
        errors : list[ErrorRecord]
            The error records to log
        """
        if not errors:
            return

        self.get_table().add([attrs.asdict(error) for error in errors])
        _LOG.info(f"Logged {len(errors)} errors in bulk")

    def get_errors(
        self,
        job_id: str | None = None,
        table_name: str | None = None,
        column_name: str | None = None,
        error_type: str | None = None,
    ) -> list[ErrorRecord]:
        """Query error records with optional filters

        Parameters
        ----------
        job_id : str, optional
            Filter by job ID
        table_name : str, optional
            Filter by table name
        column_name : str, optional
            Filter by column name
        error_type : str, optional
            Filter by error type (exception class name)

        Returns
        -------
        list[ErrorRecord]
            List of matching error records
        """
        wheres = []
        if job_id:
            wheres.append(f"job_id = '{escape_sql_string(job_id)}'")
        if table_name:
            wheres.append(f"table_name = '{escape_sql_string(table_name)}'")
        if column_name:
            wheres.append(f"column_name = '{escape_sql_string(column_name)}'")
        if error_type:
            wheres.append(f"error_type = '{escape_sql_string(error_type)}'")

        query = self.get_table(True).search()
        if wheres:
            query = query.where(" and ".join(wheres))

        # Only select known fields for forward compatibility
        known_fields = [f.name for f in attrs.fields(ErrorRecord)]
        results = query.select(known_fields).to_arrow().to_pylist()

        return [self._safe_error_record(rec) for rec in results]

    def _safe_error_record(self, rec_dict: dict) -> ErrorRecord:
        """Create ErrorRecord from dict, ignoring unknown fields"""
        known_fields = {f.name for f in attrs.fields(ErrorRecord)}
        filtered = {k: v for k, v in rec_dict.items() if k in known_fields}
        return ErrorRecord(**filtered)

    def get_failed_row_addresses(self, job_id: str, column_name: str) -> list[int]:
        """Get row addresses for all failed rows in a job

        Parameters
        ----------
        job_id : str
            Job ID to query
        column_name : str
            Column name to filter by

        Returns
        -------
        list[int]
            List of row addresses that failed
        """
        errors = self.get_errors(job_id=job_id, column_name=column_name)
        row_addresses = [
            err.row_address for err in errors if err.row_address is not None
        ]
        return row_addresses


def make_error_record_from_exception(
    exception: Exception,
    *,
    job_id: str,
    table_uri: str,
    table_name: str,
    table_version: int | None,
    column_name: str,
    udf_name: str,
    udf_version: str,
    batch_index: int,
    fragment_id: int | None = None,
    actor_id: str | None = None,
    row_address: int | None = None,
    attempt: int = 1,
    max_attempts: int = 1,
) -> ErrorRecord:
    """Factory function to create an ErrorRecord from an exception

    Parameters
    ----------
    exception : Exception
        The exception that occurred
    job_id : str
        Job ID
    table_uri : str
        URI of the table being processed
    table_name : str
        Name of the table
    table_version : int | None
        Version of the table being read
    column_name : str
        Column being computed
    udf_name : str
        Name of the UDF
    udf_version : str
        Version of the UDF
    batch_index : int
        Batch sequence number
    fragment_id : int | None, optional
        Fragment ID if applicable
    actor_id : str | None, optional
        Ray actor ID if applicable
    row_address : int | None, optional
        Row address for row-level errors
    attempt : int, optional
        Current retry attempt number
    max_attempts : int, optional
        Maximum retry attempts

    Returns
    -------
    ErrorRecord
        The constructed error record
    """
    return ErrorRecord(
        error_type=type(exception).__name__,
        error_message=str(exception),
        error_trace=traceback.format_exc(),
        job_id=job_id,
        table_uri=table_uri,
        table_name=table_name,
        table_version=table_version,
        column_name=column_name,
        udf_name=udf_name,
        udf_version=udf_version,
        actor_id=actor_id,
        fragment_id=fragment_id,
        batch_index=batch_index,
        row_address=row_address,
        attempt=attempt,
        max_attempts=max_attempts,
    )



#!/usr/bin/env python3
# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors
"""
Setup script to create the demo OpenVid table.

This script creates the initial demo table at GENEVA_DB_PATH that will be
used by e2e tests. Tests copy from this pre-existing table rather than
creating from scratch each time.

Usage:
    # Create demo table with 1000 videos (default)
    python setup_demo_table.py

    # Create demo table with custom number of videos
    python setup_demo_table.py --num-videos 500

    # Include video paths from source bucket
    python setup_demo_table.py --source-bucket gs://my-bucket/videos
"""

import argparse
import logging

from conftest import (
    DEMO_TABLE_NAME,
    GENEVA_DB_PATH,
    SOURCE_TABLE_NAME,
    VID_DATA_PATH,
    create_openvid_demo_table,
)

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s"
)
_LOG = logging.getLogger(__name__)


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Create demo OpenVid table for e2e tests"
    )
    parser.add_argument(
        "--num-videos",
        type=int,
        default=100,
        help="Number of videos to include in demo table (default: 100)",
    )
    parser.add_argument(
        "--source-bucket",
        type=str,
        default=None,
        help="GCS bucket path containing videos (e.g., gs://bucket/videos)",
    )
    args = parser.parse_args()

    _LOG.info("=" * 70)
    _LOG.info("Creating Demo OpenVid Table")
    _LOG.info("=" * 70)
    _LOG.info(f"Destination: {GENEVA_DB_PATH}")
    _LOG.info(f"Source table: {SOURCE_TABLE_NAME} (full dataset)")
    _LOG.info(f"Demo table: {DEMO_TABLE_NAME} (sample)")
    _LOG.info(f"Number of videos: {args.num_videos}")
    _LOG.info(f"Source bucket: {args.source_bucket or 'None (metadata only)'}")
    _LOG.info(f"Video data path: {VID_DATA_PATH}")
    _LOG.info("=" * 70)

    try:
        create_openvid_demo_table(
            num_videos=args.num_videos, source_bucket=args.source_bucket
        )
        _LOG.info("=" * 70)
        _LOG.info("✓ Demo tables created successfully!")
        _LOG.info(f"  Source table: {GENEVA_DB_PATH}/{SOURCE_TABLE_NAME}")
        _LOG.info(f"  Demo table: {GENEVA_DB_PATH}/{DEMO_TABLE_NAME}")
        _LOG.info(
            "  Tests will copy from demo table using openvid_table fixture"
        )
        _LOG.info("=" * 70)
    except Exception as e:
        _LOG.error("=" * 70)
        _LOG.error(f"✗ Failed to create demo table: {e}")
        _LOG.error("=" * 70)
        raise


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""
Shared pytest fixtures for integ_tests, stress_tests, and e2e_tests.

This file contains common fixtures used across all test suites.
Test-suite-specific fixtures are kept in their respective conftest.py files.
"""

import contextlib
import logging
import random
import warnings

import kubernetes
import pytest

from geneva.cluster import K8sConfigMethod
from geneva.runners.ray._mgr import ray_cluster
from geneva.runners.ray.raycluster import (
    ExitMode,
    _HeadGroupSpec,
    _WorkerGroupSpec,
)
from geneva.utils import dt_now_utc

# Try to load kubernetes config - only needed for integration/stress/e2e tests
# Unit tests don't need this, so it's okay if it fails
with contextlib.suppress(kubernetes.config.config_exception.ConfigException):
    kubernetes.config.load_kube_config()

logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")

# it's okay, we are in a test
warnings.filterwarnings(
    "ignore", "Using port forwarding for Ray cluster is not recommended for production"
)
_LOG = logging.getLogger(__name__)


# ============================================================================
# Common pytest options
# ============================================================================


def pytest_addoption(parser) -> None:
    """Add common command-line options for all test suites."""
    parser.addoption(
        "--csp",
        action="store",
        default="gcp",
        choices=["gcp", "aws"],
        help="CSP to deploy to for tests (e.g., 'gcp', 'aws')",
    )
    parser.addoption(
        "--test-slug",
        action="store",
        default=None,
        help="Test slug to identify a test run. Typically used to "
        "cleanup external resources like rayclusters",
    )
    parser.addoption(
        "--bucket-path",
        action="store",
        default=None,
        help="Bucket path to use for test data (e.g., gs://bucket/path or s3://bucket/path)",
    )
    parser.addoption(
        "--manifest",
        action="store",
        default=None,
        help="Optional custom manifest",
    )


# ============================================================================
# Common fixtures
# ============================================================================


@pytest.fixture(autouse=True, scope="session")
def geneva_k8s_service_account(csp: str) -> str:
    """
    A preconfigured service account for the test session.
    This service account should have all the permissions needed to run the tests.
    """
    return "geneva-service-account" if csp == "aws" else "geneva-integ-test"


@pytest.fixture(autouse=True, scope="session")
def geneva_test_bucket(request, slug) -> str:
    """
    Test bucket path - can be overridden with --bucket-path or defaults based on CSP.

    Falls back to default paths for local development if --bucket-path is not provided.
    """
    bucket_path = request.config.getoption("--bucket-path")

    if not bucket_path:
        csp = request.config.getoption("--csp")
        if csp == "gcp":
            bucket_path = f"gs://geneva-integ-test/{slug}/data"
        elif csp == "aws":
            bucket_path = f"s3://geneva-integ-test-devland-us-east-1/{slug}/data"
        else:
            raise ValueError(f"Unsupported --csp arg: {csp}")
        _LOG.info(f"Using default bucket path: {bucket_path}")
    else:
        _LOG.info(f"Using provided bucket path: {bucket_path}")

    return bucket_path


@pytest.fixture(scope="session")
def csp(request, geneva_test_bucket) -> str:
    """
    Cloud service provider (gcp or aws).

    Also sets up Geneva config overrides for checkpoint and upload paths
    based on geneva_test_bucket.
    """
    from geneva.config import override_config_kv

    csp = request.config.getoption("--csp")

    # Derive upload and checkpoint paths from the bucket path
    # If bucket path is: gs://bucket/slug/data
    # Then: gs://bucket/slug/data/zips and gs://bucket/slug/data/checkpoints
    override_config_kv(
        {
            "job.checkpoint.mode": "object_store",
            "uploader.upload_dir": f"{geneva_test_bucket}/zips",
            "job.checkpoint.object_store.path": f"{geneva_test_bucket}/checkpoints",
        }
    )

    return csp


@pytest.fixture(scope="session")
def manifest(request) -> str | None:
    """
    Optionally use a saved manifest instead of uploading new
    manifest
    """

    try:
        return request.config.getoption("--manifest")
    except ValueError:
        return None


@pytest.fixture(scope="session")
def slug(request) -> str | None:
    """Test slug for identifying test runs and cleanup."""
    return request.config.getoption("--test-slug") or str(random.randint(0, 10000))  # type: ignore[return-value]


@pytest.fixture(scope="session")
def region(csp) -> str:
    """Default region for the CSP."""
    return "us-east-1" if csp == "aws" else "us-central1"


@pytest.fixture(scope="session")
def k8s_config_method(csp) -> K8sConfigMethod:
    """Kubernetes config method based on CSP."""
    return K8sConfigMethod.EKS_AUTH if csp == "aws" else K8sConfigMethod.LOCAL


@pytest.fixture(scope="session")
def k8s_namespace(csp) -> str:
    """Kubernetes namespace for Ray clusters."""
    # only used for EKS auth currently
    return "geneva"


@pytest.fixture(scope="session")
def k8s_cluster_name(csp) -> str:
    """Kubernetes cluster name."""
    # only used for EKS auth currently
    return "lancedb"


@pytest.fixture(scope="session")
def head_node_selector(csp: str, node_group: str) -> dict:
    """Node selector for Ray head nodes."""
    # TODO: make node selectors consistent between EKS/GKS. This requires
    # a breaking change for GKS.  _PLACEHOLDER is odd, should make this empty dict.
    # https://linear.app/lancedb/issue/GEN-60/make-node-selectors-consistent-between-eksgks
    return (
        {"geneva.lancedb.com/ray-head": "true"}
        if csp == "aws"
        else {"_PLACEHOLDER": "true"}
    )


@pytest.fixture(scope="session")
def worker_node_selector(csp: str, node_group: str) -> dict:
    """Node selector for Ray worker nodes (CPU)."""
    # TODO: make node selectors consistent between EKS/GKS. This requires
    # a breaking change for GKS.  _PLACEHOLDER is odd, should make this empty dict.
    # https://linear.app/lancedb/issue/GEN-60/make-node-selectors-consistent-between-eksgks
    return (
        {"geneva.lancedb.com/ray-worker-cpu": "true"}
        if csp == "aws"
        else {"_PLACEHOLDER": "true"}
    )


@pytest.fixture(scope="session")
def node_group(num_gpus: int) -> str:
    """Node group name based on GPU requirements."""
    # These NGs are defined in base infra aws-enterprise eks.tf
    return (
        "lancedb-nodegroup-geneva-gpu" if num_gpus > 0 else "lancedb-nodegroup-geneva"
    )


@pytest.fixture(scope="session")
def num_gpus() -> int:
    """Number of GPUs required for tests (default: 0)."""
    return 0


@pytest.fixture
def standard_cluster(
    geneva_k8s_service_account: str,
    k8s_config_method: K8sConfigMethod,
    k8s_namespace: str,
    csp: str,
    region: str,
    head_node_selector: dict,
    worker_node_selector: dict,
    k8s_cluster_name: str,
    slug: str | None,
) -> contextlib.AbstractContextManager:
    """Standard Ray cluster for integration and e2e tests."""
    ray_cluster_name = "test-cluster"
    ray_cluster_name += f"-{dt_now_utc().strftime('%Y-%m-%d-%H-%M')}-{slug}"

    _LOG.info(f"creating ray cluster {ray_cluster_name}")

    head_spec = _HeadGroupSpec(  # type: ignore[call-arg]
        service_account=geneva_k8s_service_account,
        num_cpus=1,
        memory=3 * 1024**3,
        node_selector=head_node_selector,
    )

    # IMPORTANT: min_replicas must be ≥1 for worker groups with custom resources.
    #
    # Background: Workers with num_gpus=0 advertise a custom Ray resource via
    # rayStartParams (see raycluster.py:462-465). Each worker advertises
    # 100,000 units of the "cpu-only" custom resource to Ray.
    #
    # The Problem (Resource Discovery Bootstrapping):
    # 1. With min_replicas=0, cluster starts with zero workers
    # 2. Ray custom resources are only advertised by running workers
    # 3. When actors request cpu-only resources, Ray reports demand to autoscaler
    # 4. Kuberay autoscaler v2 cannot determine which worker group provides
    #    "cpu-only" resource because no workers are running to advertise it
    # 5. Autoscaler doesn't know which group to scale up → deadlock
    #
    # The Solution:
    # - min_replicas=1 ensures at least one worker is always running
    # - This worker advertises the cpu-only custom resource to Ray
    # - Autoscaler can now match resource demand to the correct worker group
    # - Additional workers scale up normally based on demand
    #
    # References:
    # - Ray custom resources: https://docs.ray.io/en/latest/ray-core/scheduling/resources.html
    # - Kuberay autoscaling: https://docs.ray.io/en/latest/cluster/kubernetes/user-guides/configuring-autoscaling.html
    # - Related issue: test_ray_add_column_pipeline_cpu_only_pool hung for 1.5+ hours
    #   when min_replicas=0, causing 10% of integration test failures
    worker_spec = _WorkerGroupSpec(  # type: ignore[call-arg]
        name="worker",
        min_replicas=1,
        service_account=geneva_k8s_service_account,
        num_cpus=2,
        memory=4 * 1024**3,
        node_selector=worker_node_selector,
    )

    return ray_cluster(
        name=ray_cluster_name,
        namespace=k8s_namespace,
        config_method=k8s_config_method,
        region=region,
        use_portforwarding=True,
        head_group=head_spec,
        worker_groups=[worker_spec],
        cluster_name=k8s_cluster_name,
        role_name="geneva-client-role",
        on_exit=ExitMode.DELETE,
        extra_env={
            "RAY_BACKEND_LOG_LEVEL": "info",
            "RAY_LOG_TO_DRIVER": "1",
            "RAY_ENABLE_RECORD_ACTOR_TASK_LOGGING": "1",
            "RAY_RUNTIME_ENV_LOG_TO_DRIVER_ENABLED": "true",
        },
        log_to_driver=True,
        logging_level=logging.INFO,
    )


@pytest.fixture
def beefy_cluster(
    geneva_k8s_service_account: str,
    k8s_config_method: K8sConfigMethod,
    k8s_namespace: str,
    csp: str,
    region: str,
    head_node_selector: dict,
    worker_node_selector: dict,
    k8s_cluster_name: str,
    slug: str | None,
) -> contextlib.AbstractContextManager:
    """Large Ray cluster for stress/e2e tests with high CPU/memory workers."""
    ray_cluster_name = "beefy-cluster"
    ray_cluster_name += f"-{dt_now_utc().strftime('%Y-%m-%d-%H-%M')}-{slug}"

    _LOG.info(f"creating beefy ray cluster {ray_cluster_name}")

    head_spec = _HeadGroupSpec(  # type: ignore[call-arg]
        service_account=geneva_k8s_service_account,
        num_cpus=1,
        memory=3 * 1024**3,
        node_selector=head_node_selector,
    )

    worker_spec = _WorkerGroupSpec(  # type: ignore[call-arg]
        name="worker",
        min_replicas=0,
        service_account=geneva_k8s_service_account,
        num_cpus=14,
        memory=56 * 1024**3,
        node_selector=worker_node_selector,
        env_vars={
            "LANCE_IO_THREADS": "4",
            "LANCE_PROCESS_IO_THREADS_LIMIT": "8",
        },
    )

    return ray_cluster(
        name=ray_cluster_name,
        namespace=k8s_namespace,
        config_method=k8s_config_method,
        region=region,
        use_portforwarding=True,
        head_group=head_spec,
        worker_groups=[worker_spec],
        cluster_name=k8s_cluster_name,
        role_name="geneva-client-role",
        on_exit=ExitMode.DELETE,
    )



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# module for packaging workspace and UDFs
# also data spec for persisting the artifacts

import abc
import base64
import hashlib
import json
import logging
from pathlib import Path

import attrs
from pyarrow.fs import FileSystem, FileType
from typing_extensions import Self

import geneva.cloudpickle as cloudpickle
from geneva.config import ConfigBase
from geneva.packager.zip import WorkspaceZipper
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


class UDFBackend(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def packager(cls) -> "UDFPackager":
        """Return the packager for this backend."""

    def to_bytes(self) -> bytes:
        return json.dumps(attrs.asdict(self)).encode()

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        return cls(**json.loads(data.decode()))


@attrs.define
class UDFSpec:
    """Specification for a user-defined function.

    This is an holder of an arbitrary user-defined function,
    which can use an backend for marshalling.

    The most common is likely Docker + some kind of workspace
    persistence. However, we want to support more than just
    Docker, so we create this "out most" abstraction to allow
    for more flexibility.
    """

    # the name of the udf
    name: str = attrs.field()

    def __attrs_post_init__(self) -> None:
        if not self.name:
            raise ValueError("UDF name must not be empty.")
        if len(self.name) < 1:
            raise ValueError("UDF name must be at least 1 character long.")

        backend_names = [cls.__name__ for cls in UDFBackend.__subclasses__()]
        unique_backend_names = set(backend_names)
        if self.backend not in unique_backend_names:
            raise ValueError(f"Unknown backend: {self.backend}")

    # the packaging backend for the udf
    backend: str = attrs.field()

    udf_payload: bytes = attrs.field()

    # the payload for the runner -- This is a HACK for allowing phalanx knowing
    # how to dispatch the UDF job. Make sure changes here are compatible to
    # parsing in phalanx.
    runner_payload: bytes | None = attrs.field(default=None)

    @classmethod
    def udf_from_spec(cls, data) -> UDF:
        # TODO: load the spec and find the backend,
        # then call the packager to do the next level unmarshalling
        raise NotImplementedError("udf_from_spec not yet implemented")


@attrs.define
class DockerUDFSpecV1(UDFBackend):
    """Specification for a user-defined function that runs in a Docker container.
    -- Version 1

    In this packaging spec, the python interpreter is assumed to be correctly
    setup in the container, and the user-defined function is expected to load
    using cloudpickle. With the option of downloading additional workspace
    files from a remote location (S3, GCS, etc).
    """

    # the image to run the udf in
    image: str = attrs.field()

    # the tag of the image
    tag: str | None = attrs.field()

    # optionally have a zip of the workspace and store it separately
    # this should be the path to the zip file on (S3, GCS, etc)
    workspace_zip: str | None = attrs.field()

    # the checksum of the workspace zip
    workspace_checksum: str | None = attrs.field()

    # the udf pickle to run
    udf_pickle: bytes = attrs.field()

    # the checksum of the udf pickle
    udf_checksum: str = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        # Validate tag
        if self.tag is not None:
            if len(self.tag) > 128:
                raise ValueError("Tag must be less than 128 characters.")
            if not all(
                c.isalpha() or c.isnumeric() or c in {"_", ".", "-"} for c in self.tag
            ):
                raise ValueError("Tag must be valid alphanumeric.")

        # Validate workspace checksum
        if self.workspace_zip and not self.workspace_checksum:
            raise ValueError(
                "Workspace checksum must not be empty when a workspace is provided."
            )

        # Validate UDF pickle
        if not self.udf_pickle:
            raise ValueError("UDF pickle must not be empty.")

        # Try to validate the UDF by unpickling, but don't fail if modules are missing
        # This supports distributed workflows where manifests are created in one
        # environment and used in another (e.g., uploaded by CI, used in notebooks)
        try:
            udf = cloudpickle.loads(self.udf_pickle)
            if not isinstance(udf, UDF):
                raise ValueError("UDF pickle must contain a UDF object.")
        except ModuleNotFoundError as e:
            _LOG.warning(
                f"Could not validate UDF pickle during spec loading: {e}. "
                "This is expected if the UDF was created in a different environment. "
                "Validation will happen when the UDF is executed on Ray workers."
            )

        self.udf_checksum = hashlib.sha256(self.udf_pickle).hexdigest()

    @classmethod
    def packager(cls) -> "UDFPackager":
        return DockerUDFPackager()

    def to_bytes(self) -> bytes:
        self_as_dict = attrs.asdict(self)
        self_as_dict["udf_pickle"] = base64.b64encode(
            self_as_dict["udf_pickle"]
        ).decode("utf-8")
        return json.dumps(self_as_dict).encode()

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        self_as_dict = json.loads(data.decode())
        self_as_dict["udf_pickle"] = base64.b64decode(
            self_as_dict["udf_pickle"].encode("utf-8")
        )

        checksum = self_as_dict.pop("udf_checksum")  # not part of the init
        val = cls(**self_as_dict)
        val.udf_checksum = checksum

        return val


class UDFPackager(abc.ABC):
    """Packager for user-defined functions."""

    @abc.abstractmethod
    def marshal(self, udf: UDF) -> UDFSpec:
        """Marshal a user-defined function."""

    @abc.abstractmethod
    def unmarshal(self, spec: UDFSpec) -> UDF | None:
        """Unmarshal a user-defined function.

        Returns None if the UDF cannot be unpickled due to missing modules.
        """


@attrs.define
class _DockerUDFPackagerConfig(ConfigBase):
    prebuilt_docker_img: str | None = attrs.field(default=None)

    # the backend the image will eventually run on. Gets passed ot the
    # docker workspace packager it can know which base image/dockerfile
    # template to use
    runtime_backend: str | None = attrs.field(default=None)

    workspace_upload_location: str | None = attrs.field(default=None)

    @classmethod
    def name(cls) -> str:
        return "docker"


@attrs.define
class _UDFPackagerConfig(ConfigBase):
    docker: _DockerUDFPackagerConfig = attrs.field(default=_DockerUDFPackagerConfig())

    @classmethod
    def name(cls) -> str:
        return "udf"


@attrs.define
class DockerUDFPackager(UDFPackager):
    # If the user wants to use an prebuilt docker image, they can provide the
    # image name:tag here. This will be used instead of building and pushing a
    # new image.
    prebuilt_docker_img: str | None = attrs.field(default=None)

    # the location to upload the zipped workspace to
    # this should be the path to some directory on object storage (S3, GCS, etc)
    workspace_upload_location: str | None = attrs.field(default=None)

    # provide a zipper implementation with the correct configuration for how to
    # to zip the workspace. If workspace_zip is supplied and this is not, then
    # the packager will use the default WorkspaceZipper.
    zip_workspace_packager: WorkspaceZipper | None = attrs.field(default=None)

    def __attrs_post_init__(self) -> None:
        # Set default prebuilt_docker_img
        if self.prebuilt_docker_img is None:
            config = _UDFPackagerConfig.get()
            if config.docker is not None:
                self.prebuilt_docker_img = config.docker.prebuilt_docker_img

        # Set default workspace_upload_location
        if self.workspace_upload_location is None:
            config = _UDFPackagerConfig.get()
            if config.docker is not None:
                self.workspace_upload_location = config.docker.workspace_upload_location

        # Set default zip_workspace_packager
        if self.zip_workspace_packager is None and self.workspace_upload_location:
            self.zip_workspace_packager = WorkspaceZipper(path=Path("."))

    def marshal(self, udf: UDF) -> UDFSpec:
        image_name, tag = "test-image", "latest"

        workspace_zip = None
        workspace_checksum = None
        if self.zip_workspace_packager:
            _LOG.info("Packaging zipped workspace")
            zip_path, checksum = self.zip_workspace_packager.zip()
            _LOG.info("Uploading zipped workspace")

            upload_location = self.workspace_upload_location
            if upload_location is not None and upload_location[-1] != "/":
                upload_location += "/"

            if upload_location is None:
                raise ValueError("workspace_upload_location cannot be None")
            remote_fs, root_path = FileSystem.from_uri(upload_location)
            out_path = f"{root_path}/{checksum}.zip"
            curr_remote_file_info = remote_fs.get_file_info(out_path)
            if curr_remote_file_info.type == FileType.NotFound:
                _LOG.info(
                    f"Workspace zip does not exist, uploading {zip_path} to {out_path}"
                )
                if isinstance(zip_path, list):
                    raise ValueError("zip_path should be a single Path, not a list")
                local_fs, _local_root = FileSystem.from_uri(
                    zip_path.absolute().parent.as_uri()
                )

                with (
                    local_fs.open_input_stream(str(zip_path)) as in_file,
                    remote_fs.open_output_stream(out_path) as out_file,
                ):
                    bath_size = 1024 * 1024
                    while True:
                        buf = in_file.read(bath_size)
                        if buf:
                            out_file.write(buf)
                        else:
                            break

                _LOG.info(f"Uploaded workspace zip to {out_path}")
            else:
                _LOG.info("Workspace zip already exists, skipping upload")

            workspace_zip = f"{upload_location}{checksum}.zip"
            workspace_checksum = checksum

        udf_pickle = cloudpickle.dumps(udf)

        return UDFSpec(
            name=udf.name,
            backend=DockerUDFSpecV1.__name__,
            udf_payload=DockerUDFSpecV1(
                image=image_name,
                tag=tag,
                workspace_zip=workspace_zip,
                workspace_checksum=workspace_checksum,
                udf_pickle=udf_pickle,
            ).to_bytes(),
            runner_payload=json.dumps(
                {
                    "image": image_name + ":" + tag,
                }
            ).encode(),
        )

    def unmarshal(self, spec: UDFSpec) -> UDF | None:
        """Unmarshal a UDF from a spec.

        Returns None if the UDF cannot be unpickled due to missing modules.
        This supports distributed workflows where manifests are created in one
        environment and used in another.
        """
        docker_spec = self.backend(spec)
        try:
            udf = cloudpickle.loads(docker_spec.udf_pickle)
            if not isinstance(udf, UDF):
                raise ValueError("UDF pickle must contain a UDF object.")
            return udf
        except ModuleNotFoundError as e:
            _LOG.warning(
                f"Cannot unmarshal UDF for validation: {e}. "
                "This is expected if the UDF was created in a different environment. "
                "Skipping client-side validation. The UDF will be executed on Ray "
                "workers where modules are available via py_modules in the manifest."
            )
            return None

    def backend(self, spec: UDFSpec) -> DockerUDFSpecV1:
        if spec.backend != DockerUDFSpecV1.__name__:
            raise ValueError("Invalid backend for UDF spec.")

        return DockerUDFSpecV1.from_bytes(spec.udf_payload)



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import logging
import time
from datetime import timedelta
from pathlib import Path
from typing import Any, NamedTuple

import lance
import pyarrow as pa
import pytest
import ray

import geneva
from geneva import udf
from geneva.db import Connection
from geneva.runners.ray.pipeline import RayJobFuture
from geneva.table import Table

_LOG = logging.getLogger(__name__)
_LOG.setLevel(logging.DEBUG)

SIZE = 17  # was 256
MAX_ROWS_PER_FILE = 5


@pytest.fixture(autouse=True, scope="class")
def ray_cluster() -> None:
    ray.shutdown()
    ray.init(
        log_to_driver=True,
        logging_config=ray.LoggingConfig(
            encoding="TEXT", log_level="INFO", additional_log_standard_attrs=["name"]
        ),
    )
    yield
    ray.shutdown()


@pytest.fixture(autouse=True)
def db(tmp_path) -> Connection:
    tbl_path = tmp_path / "foo.lance"
    make_new_ds_a(tbl_path)
    db = geneva.connect(str(tmp_path), read_consistency_interval=timedelta(0))
    yield db
    db.close()


@pytest.fixture
def tbl_path(tmp_path) -> Path:
    return tmp_path / "foo.lance"


def make_new_ds_a(tbl_path: Path) -> lance.dataset:
    # create initial dataset with only column 'a'
    data = {"a": pa.array(range(SIZE))}
    tbl = pa.Table.from_pydict(data)
    ds = lance.write_dataset(tbl, tbl_path, max_rows_per_file=MAX_ROWS_PER_FILE)
    return ds


class UDFTestConfig(NamedTuple):
    expected_recordbatch: dict[Any, Any]
    where: str | None = None


def int32_return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
    return pa.RecordBatch.from_pydict(
        {"b": pa.array([None] * batch.num_rows, pa.int32())}
    )


def setup_table_and_udf_column(
    db: Connection,
    shuffle_config,
    udf,
) -> Table:
    tbl = db.open_table("foo")

    tbl.add_columns(
        {"b": udf},
        **shuffle_config,
    )
    _LOG.info(f"Table prebackfill at version {tbl.version}")
    return tbl


def verify_backfill(tbl: Table, col, fut) -> None:
    tbl.checkout_latest()
    _LOG.info(f"completed backfill, now on version {tbl.version}")

    res = tbl.search().select(["a", col]).to_arrow().to_pydict()

    expected = expected_res(col)
    _LOG.info(f"actual={res} expected={expected}")
    assert res == expected, "wrong results"

    # Verify job history record seems reasonable
    hist = tbl._conn._history
    jr = hist.get(fut.job_id)[0]
    _LOG.info(f"{jr=}")

    assert jr.status == "DONE"
    assert jr.object_ref is not None
    assert jr.table_name == tbl.name
    assert jr.column_name == col
    assert jr.launched_at is not None
    assert jr.completed_at is not None

    # Verify progress tracking in future seems right.
    assert fut.job_id is not None
    assert isinstance(fut, RayJobFuture)
    assert fut.ray_obj_ref is not None

    _LOG.info(f"{fut._pbars=}")

    assert fut._pbars["rows_checkpointed"].n == fut._pbars["rows_checkpointed"].total
    assert (
        fut._pbars["rows_ready_for_commit"].n
        == fut._pbars["rows_ready_for_commit"].total
    )
    assert fut._pbars["rows_committed"].n == fut._pbars["rows_committed"].total


# 0.1 cpu so we don't wait for provisioning in the tests
@udf(data_type=pa.int32(), batch_size=8, num_cpus=1)
def times_ten(a) -> int:
    time.sleep(0.1)  # simulate some work
    return a * 10


def expected_res(col: str) -> dict[str, list[int]]:
    return {
        "a": list(range(SIZE)),
        col: [x * 10 for x in range(SIZE)],
    }


default_shuffle_config = {
    "batch_size": 1,
    "shuffle_buffer_size": 3,
    "task_shuffle_diversity": None,
}


@pytest.mark.multibackfill
@pytest.mark.ray
def test_pipeline_concurrent(db: Connection) -> None:
    tbl = db.open_table("foo")

    cols = ["b", "c", "d", "e", "f", "g", "h", "i", "j"]
    adds = dict.fromkeys(cols, times_ten)
    tbl.add_columns(
        adds,
        **default_shuffle_config,
    )

    futs = {}
    for col in cols:
        _LOG.info(f"Starting column {col} backfill")
        futs[col] = tbl.backfill_async(col, where=None)

    for col, fut in futs.items():
        _LOG.info(f"Verifying backfill {col} with {fut}...")
        while not fut.done():
            _LOG.info(f"Waiting for backfill {col} to complete...")

        _LOG.info(f"current results: {tbl.to_arrow().to_pydict()}")
        verify_backfill(tbl, col, fut)



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import base64
import logging
import tempfile
from datetime import datetime, timedelta

from kubernetes import client, config

from geneva.cluster import K8sConfigMethod

_LOG = logging.getLogger(__name__)

# Default EKS token expiration period in seconds (30mins)
TOKEN_EXPIRATION_S = 1800


def build_api_client(
    config_method: K8sConfigMethod,
    region: str | None,
    cluster_name: str | None,
    role_name: str | None,
) -> client.ApiClient | None:
    """Build the k8s API client based on the configuration method and region.
    Returns None for default config method (non-EKS).
    role_name can be an ARN or role name in current account.
    """

    if config_method == K8sConfigMethod.IN_CLUSTER:
        config.load_incluster_config()

        return None
    elif config_method == K8sConfigMethod.LOCAL:
        config.load_kube_config()

        return None
    elif config_method == K8sConfigMethod.EKS_AUTH:
        ca_data, endpoint, token = _eks_auth(cluster_name, region, role_name)
        cafile = _write_cafile(ca_data)
        c = _api_client(endpoint, token, cafile.name)
        return c
    else:
        raise Exception(f"unsupported config method: {config_method}")


def _eks_auth(
    cluster_name: str | None, region: str | None, role_name: str | None
) -> tuple[str, str, str]:
    role_name = role_name or "geneva-client-role"
    region = region or "us-east-1"
    cluster_name = cluster_name or "lancedb"

    _LOG.debug(f"authenticating with EKS. {role_name=}, {region=}, {cluster_name=}")
    # this role requires an EKS Access Entry with
    # AmazonEKSClusterAdminPolicy on the given namespace
    # and must be assumable by the current principal
    role_arn = _get_role_arn(region, role_name)

    # Get AWS session with assumed role credentials (if role provided)
    aws_session = _get_aws_session_with_role(region, role_arn)
    token = _get_token(cluster_name, role_arn, region, TOKEN_EXPIRATION_S)["status"][
        "token"
    ]

    eks_client = aws_session.client("eks")
    cluster_data = eks_client.describe_cluster(name=cluster_name)["cluster"]
    ca_data = cluster_data["certificateAuthority"]["data"]
    endpoint = cluster_data["endpoint"]
    _LOG.info(f"authenticated with EKS. {cluster_name=} {role_arn=}")

    return ca_data, endpoint, token


def _get_role_arn(region: str, role_name: str | None) -> str | None:
    import boto3

    # default to using current credentials
    role_arn = None
    if role_name:
        # use role ARN if provided
        if role_name.startswith("arn:aws:iam::"):
            role_arn = role_name
        else:
            # if role name is provided, assume it in the current account
            identity = boto3.client("sts", region_name=region).get_caller_identity()
            _LOG.debug("caller identity: %s", identity)
            acct_id = identity.get("Account")
            role_arn = f"arn:aws:iam::{acct_id}:role/{role_name}"

    return role_arn


def _write_cafile(data: str) -> tempfile.NamedTemporaryFile:  # type: ignore[type-arg]
    # ruff: noqa: SIM115
    cafile = tempfile.NamedTemporaryFile(mode="wb", delete=False)
    cadata_b64 = data
    cadata = base64.b64decode(cadata_b64)
    cafile.write(cadata)
    cafile.flush()
    return cafile


def _api_client(endpoint: str, token: str, cafile: str) -> client.ApiClient:
    kconfig = config.kube_config.Configuration(
        host=endpoint, api_key={"authorization": "Bearer " + token}
    )
    # Set ssl_ca_cert only if the configuration supports it
    if hasattr(kconfig, "ssl_ca_cert"):
        kconfig.ssl_ca_cert = cafile  # type: ignore[assignment]
    kclient = client.ApiClient(configuration=kconfig)
    return kclient


def _get_expiration_time(token_expiration_s: int) -> str:
    token_expiration = datetime.utcnow() + timedelta(seconds=token_expiration_s)
    return token_expiration.strftime("%Y-%m-%dT%H:%M:%SZ")


def _get_exp_s() -> int:
    return TOKEN_EXPIRATION_S


def _get_token(
    cluster_name: str,
    role_arn: str | None,
    region_name: str,
    token_expiration_s: int = _get_exp_s(),
) -> dict:
    from awscli.customizations.eks.get_token import (
        STSClientFactory,
        TokenGenerator,
    )
    from botocore import session

    work_session = session.get_session()
    client_factory = STSClientFactory(work_session)
    sts_client = client_factory.get_sts_client(
        role_arn=role_arn, region_name=region_name
    )

    token = TokenGenerator(sts_client).get_token(cluster_name)
    exp = _get_expiration_time(token_expiration_s)
    _LOG.debug(f"EKS token expiration: {exp}")
    return {
        "kind": "ExecCredential",
        "apiVersion": "client.authentication.k8s.io/v1beta1",
        "spec": {},
        "status": {
            "expirationTimestamp": exp,
            "token": token,
        },
    }


def _get_aws_session_with_role(region_name: str, role_arn: str | None):  # noqa: ANN202
    """Get AWS session with assumed role if role_arn provided,
    otherwise default session"""
    import boto3

    if role_arn:
        # Assume the role to get temporary credentials
        sts = boto3.client("sts", region_name=region_name)
        assumed_role = sts.assume_role(
            RoleArn=role_arn, RoleSessionName="geneva-session"
        )
        credentials = assumed_role["Credentials"]

        # Create session with assumed role credentials
        return boto3.Session(
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"],
            region_name=region_name,
        )
    else:
        # Use default session when no role is specified
        return boto3.Session(region_name=region_name)



# ruff: noqa: PERF203

# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# super simple applier

import logging

import attrs
import pyarrow as pa

from geneva.apply.applier import BatchApplier
from geneva.apply.error_handling import BatchStrategy, ErrorHandlingContext
from geneva.apply.task import MapTask, ReadTask
from geneva.debug.logger import ErrorLogger

_LOG = logging.getLogger(__name__)


@attrs.define
class SimpleApplier(BatchApplier):
    """
    A simple applier that applies a function to each element in the batch.
    """

    job_id: str = attrs.field(default="unknown")

    def run(
        self,
        read_task: ReadTask,
        map_task: MapTask,
        error_logger: ErrorLogger,
    ) -> pa.RecordBatch:
        batches = []
        for seq, batch in enumerate(
            read_task.to_batches(batch_size=map_task.batch_size())
        ):
            # Create error context and strategy for this batch
            ctx = ErrorHandlingContext.create(map_task, read_task, self.job_id, seq)
            strategy = BatchStrategy.from_context(ctx, map_task, error_logger)

            # Apply strategy - no branching needed!
            result_batch, error_records = strategy.apply(batch)

            # Log errors if any were collected
            if error_records:
                error_logger.log_errors(error_records)

            batches.append(result_batch)

        if not batches:
            return pa.RecordBatch.from_arrays([])

        schema = batches[0].schema
        combined_table = pa.Table.from_batches(batches, schema)
        combined_table = combined_table.combine_chunks()
        combined_batch = combined_table.to_batches()[0]

        _LOG.info(
            f"Simple applier: completed {map_task.name()} for {read_task} with "
            f"{len(combined_batch)} rows"
        )
        return combined_batch



#!/usr/bin/env python
# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""
Upload manifest and add columns for sentence transformer UDF to Geneva.

This script runs in the UDF's dependency environment (via uv run)
and uploads the manifest to the Geneva connection, then adds columns
to the specified table.

Usage:
    export GENEVA_TABLE_NAME=oxford_pets_shared_abc123
    cd e2e/oxford-pets/udfs/sentence-transformers
    uv run python upload_manifest.py --bucket gs://bucket/path

Environment Variables:
    GENEVA_TABLE_NAME: Table name to add columns to (required)
    IMAGE_ID_EMBEDDING_COL: Custom column name for image_id embeddings (default: "image_id_embedding")
"""

import logging
import os

from geneva.manifest.uploader_cli import main_upload_workflow
from manifest import create_manifest

logging.basicConfig(level=logging.INFO)
_LOG = logging.getLogger(__name__)

from geneva.udfs.embeddings import sentence_transformer_udf

# Create the UDF that embeds text from image_id field
# Using dimension=384 for lazy loading (all-MiniLM-L6-v2 model)
text_embedding = sentence_transformer_udf(
    model="sentence-transformers/all-MiniLM-L6-v2",
    column="image_id",  # Use image_id as input text
    normalize=True,
    trust_remote_code=True,
    dimension=384,  # Specify dimension to enable lazy loading
)


def get_columns() -> dict[str, callable]:
    """
    Get column definitions for this UDF package.

    Column names can be customized via environment variables:
    - IMAGE_ID_EMBEDDING_COL: Custom name for text embedding column
      (default: "image_id_embedding")

    Returns:
        Dictionary mapping column names to UDF functions
    """
    return {
        os.getenv("IMAGE_ID_EMBEDDING_COL", "image_id_embedding"): text_embedding,
    }


if __name__ == "__main__":
    main_upload_workflow(
        get_columns_fn=get_columns,
        create_manifest_fn=create_manifest,
        description="Upload sentence transformer UDF manifest and add columns to table",
        default_manifest_name="sentence-transformers-udfs-v1",
        logger=_LOG,
    )



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

"""
E2E test-specific fixtures for Oxford Pets suite.

This conftest is fully self-contained and does not depend on src/conftest.py.
All fixtures needed for e2e tests are defined here.
"""

import contextlib
import logging
import os
import random
import sys
import uuid
import warnings
from collections.abc import Generator
from pathlib import Path

import kubernetes
import pyarrow as pa
import pytest

from geneva.cluster import GenevaClusterType, K8sConfigMethod
from geneva.cluster.mgr import GenevaCluster, HeadGroupConfig, KubeRayConfig, WorkerGroupConfig
from geneva.utils import dt_now_utc

# Add UDF directories to sys.path so UDF modules can be imported during test validation
# This is needed because tests unpickle UDFs for validation before sending to Ray
_udfs_dir = Path(__file__).parent / "udfs"
for udf_package in ["simple", "openclip", "blip"]:
    udf_path = _udfs_dir / udf_package
    if udf_path.exists() and str(udf_path) not in sys.path:
        sys.path.insert(0, str(udf_path))

# Try to load kubernetes config - only needed for e2e tests
with contextlib.suppress(kubernetes.config.config_exception.ConfigException):
    kubernetes.config.load_kube_config()

logging.basicConfig(level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")

warnings.filterwarnings(
    "ignore", "Using port forwarding for Ray cluster is not recommended for production"
)

_LOG = logging.getLogger(__name__)

# Flag to track if manifests have been uploaded
_MANIFESTS_UPLOADED = False


def _upload_all_manifests(bucket_path: str) -> None:
    """
    Upload all UDF manifests and add columns to the table.

    This runs after the table is created but before tests execute.
    Each upload script runs in its own UDF environment via `uv run`.
    """
    import subprocess
    from pathlib import Path

    _LOG.info("Uploading UDF manifests and adding columns...")

    udfs_dir = Path(__file__).parent / "udfs"
    udf_packages = ["simple", "openclip", "blip", "sentence-transformers"]

    for udf_name in udf_packages:
        udf_dir = udfs_dir / udf_name
        upload_script = udf_dir / "upload_manifest.py"

        _LOG.info(f"Uploading manifest for '{udf_name}' UDF package...")

        # Run upload script in UDF's environment
        # Step 1: Sync dependencies to populate .venv with UDF dependencies
        # (includes geneva and google-cloud-storage from pyproject.toml)
        sync_result = subprocess.run(
            ["uv", "sync"],
            cwd=str(udf_dir),
            capture_output=True,
            text=True,
        )
        if sync_result.returncode != 0:
            _LOG.error(f"Failed to sync dependencies for '{udf_name}':")
            if sync_result.stderr:
                for line in sync_result.stderr.splitlines():
                    _LOG.error(f"  [{udf_name}] {line}")
            raise RuntimeError(f"Dependency sync failed for '{udf_name}'")

        # Step 2: Run upload script from UDF's .venv
        result = subprocess.run(
            ["uv", "run", "python", "upload_manifest.py", "--bucket", bucket_path],
            cwd=str(udf_dir),
            capture_output=True,
            text=True,
            env=os.environ.copy(),
        )

        # Always show the output for debugging
        if result.stdout:
            for line in result.stdout.splitlines():
                _LOG.info(f"  [{udf_name}] {line}")
        if result.stderr:
            for line in result.stderr.splitlines():
                _LOG.warning(f"  [{udf_name}] {line}")

        if result.returncode != 0:
            _LOG.error(f"Failed to upload manifest for '{udf_name}':")
            raise RuntimeError(f"Manifest upload failed for '{udf_name}'")

        _LOG.info(f"✓ Manifest '{udf_name}' uploaded successfully")

    _LOG.info("All manifests uploaded and columns added")


# ============================================================================
# E2E test-specific pytest options
# ============================================================================


def pytest_addoption(parser) -> None:
    """Add e2e-specific command-line options."""
    parser.addoption(
        "--csp",
        action="store",
        default="gcp",
        choices=["gcp", "aws"],
        help="CSP to deploy to for tests (e.g., 'gcp', 'aws')",
    )
    parser.addoption(
        "--test-slug",
        action="store",
        default=None,
        help="Test slug to identify a test run",
    )
    parser.addoption(
        "--bucket-path",
        action="store",
        default=None,
        help="Bucket path for test data (e.g., gs://bucket/path or s3://bucket/path)",
    )
    parser.addoption(
        "--num-images",
        action="store",
        type=int,
        default=500,
        help="Number of images to process from Oxford pets dataset",
    )
    parser.addoption(
        "--batch-size",
        action="store",
        type=int,
        default=10,
        help="Batch size for backfill operations",
    )
    parser.addoption(
        "--skip-gpu",
        action="store_true",
        default=False,
        help="Skip GPU-based tests (captions, GPU embeddings)",
    )


# ============================================================================
# Common fixtures (copied from src/conftest.py for independence)
# ============================================================================


@pytest.fixture(scope="session")
def csp(request) -> str:
    """Cloud service provider (gcp or aws)."""
    return request.config.getoption("--csp")


@pytest.fixture(scope="session")
def slug(request) -> str:
    """Test slug for identifying test runs and cleanup."""
    return request.config.getoption("--test-slug") or str(random.randint(0, 10000))


@pytest.fixture(scope="session")
def geneva_test_bucket(request, slug, csp) -> str:
    """
    Test bucket path - can be overridden with --bucket-path or defaults based on CSP.

    Also sets up Geneva config overrides for checkpoint and upload paths.
    """
    from geneva.config import override_config_kv

    bucket_path = request.config.getoption("--bucket-path")

    if not bucket_path:
        if csp == "gcp":
            bucket_path = f"gs://geneva-integ-test/{slug}/data"
        elif csp == "aws":
            bucket_path = f"s3://geneva-integ-test-devland-us-east-1/{slug}/data"
        else:
            raise ValueError(f"Unsupported --csp arg: {csp}")
        _LOG.info(f"Using default bucket path: {bucket_path}")
    else:
        _LOG.info(f"Using provided bucket path: {bucket_path}")

    # Set Geneva config overrides
    override_config_kv(
        {
            "job.checkpoint.mode": "object_store",
            "uploader.upload_dir": f"{bucket_path}/zips",
            "job.checkpoint.object_store.path": f"{bucket_path}/checkpoints",
        }
    )

    return bucket_path


@pytest.fixture(scope="session")
def geneva_k8s_service_account(csp: str) -> str:
    """Preconfigured service account for the test session."""
    return "geneva-service-account" if csp == "aws" else "geneva-integ-test"


@pytest.fixture(scope="session")
def region(csp: str) -> str:
    """Default region for the CSP."""
    return "us-east-1" if csp == "aws" else "us-central1"


@pytest.fixture(scope="session")
def k8s_config_method(csp: str) -> K8sConfigMethod:
    """Kubernetes config method based on CSP."""
    return K8sConfigMethod.EKS_AUTH if csp == "aws" else K8sConfigMethod.LOCAL


@pytest.fixture(scope="session")
def k8s_namespace(csp: str) -> str:
    """Kubernetes namespace for Ray clusters."""
    return "geneva"


@pytest.fixture(scope="session")
def k8s_cluster_name(csp: str) -> str:
    """Kubernetes cluster name."""
    return "lancedb"


@pytest.fixture(scope="session")
def head_node_selector(csp: str) -> dict:
    """Node selector for Ray head nodes."""
    return (
        {"geneva.lancedb.com/ray-head": "true"}
        if csp == "aws"
        else {"_PLACEHOLDER": "true"}
    )


@pytest.fixture(scope="session")
def worker_node_selector(csp: str) -> dict:
    """Node selector for Ray worker nodes (CPU)."""
    return (
        {"geneva.lancedb.com/ray-worker-cpu": "true"}
        if csp == "aws"
        else {"_PLACEHOLDER": "true"}
    )


# ============================================================================
# Dataset Loading Utilities
# ============================================================================

ImageBatchGenerator = Generator[pa.RecordBatch, None, None]


def load_oxford_pets_images(
    num_images: int = 500, frag_size: int = 25
) -> ImageBatchGenerator:
    """
    Load images from the Oxford-IIIT Pet dataset.

    Args:
        num_images: Number of images to load from the dataset
        frag_size: Number of images per fragment

    Yields:
        PyArrow RecordBatch with columns: image (bytes), label (string)

    Raises:
        pytest.skip: If dataset cannot be loaded due to network or API errors
    """
    import io

    import pyarrow as pa
    from datasets import load_dataset

    from geneva.tqdm import tqdm

    _LOG.info(f"Loading {num_images} images from Oxford pets dataset")

    try:
        # there are 3680 images.  If num_images > 3680, it will just load all
        dataset = load_dataset("timm/oxford-iiit-pet", split=f"train[:{num_images}]")
    except Exception as e:
        pytest.skip(
            f"Failed to load Oxford pets dataset from HuggingFace. "
            f"This may be due to network issues or API unavailability. Error: {e}"
        )

    batch = []
    for row in tqdm(dataset):
        buf = io.BytesIO()
        row["image"].save(buf, format="png")
        batch.append({"image": buf.getvalue(), "label": row["label"],
            "image_id":row["image_id"]})

        if len(batch) >= frag_size:
            yield pa.RecordBatch.from_pylist(batch)
            batch = []

    if batch:
        yield pa.RecordBatch.from_pylist(batch)


# ============================================================================
# E2E test-specific fixtures
# ============================================================================


@pytest.fixture(scope="session")
def num_images(request) -> int:
    """Number of images to process in e2e tests."""
    return request.config.getoption("--num-images")


@pytest.fixture(scope="session")
def batch_size(request) -> int:
    """Batch size for backfill operations in e2e tests."""
    return request.config.getoption("--batch-size")


@pytest.fixture(scope="session")
def skip_gpu(request) -> bool:
    """Whether to skip GPU-based tests."""
    return request.config.getoption("--skip-gpu")


@pytest.fixture(scope="session")
def oxford_pets_table(geneva_test_bucket: str, num_images: int) -> tuple:  # type: ignore[misc]
    """
    Session-scoped fixture that creates a shared table with Oxford pets images.

    This table is created once per test session and reused across all e2e tests,
    avoiding repeated dataset downloads.

    Also sets GENEVA_TABLE_NAME environment variable for upload scripts.

    Returns:
        tuple: (connection, table, table_name)
    """
    import geneva

    _LOG.info(f"Creating shared Oxford pets table with {num_images} images")

    conn = geneva.connect(geneva_test_bucket)
    table_name = f"oxford_pets_shared_{uuid.uuid4().hex}"

    # Load images and create table (only happens once per session)
    first = True
    for batch in load_oxford_pets_images(num_images):
        if first:
            tbl = conn.create_table(table_name, batch, mode="overwrite")
            first = False
        else:
            tbl.add(batch)

    _LOG.info(
        f"Shared table created: name='{table_name}', rows={len(tbl)}, "
        f"schema={tbl.schema}. This will be reused across all e2e tests."
    )

    # Export table name as environment variable for upload scripts
    os.environ["GENEVA_TABLE_NAME"] = table_name
    _LOG.info(f"Set GENEVA_TABLE_NAME={table_name}")

    # Upload manifests and add columns (once per session)
    global _MANIFESTS_UPLOADED
    if not _MANIFESTS_UPLOADED:
        _upload_all_manifests(geneva_test_bucket)
        _MANIFESTS_UPLOADED = True

        # Refresh table to pick up newly added columns
        tbl = conn.open_table(table_name)
        _LOG.info(f"Table schema after manifest uploads: {tbl.schema}")

    yield conn, tbl, table_name


@pytest.fixture
def standard_cluster(
    oxford_pets_table: tuple,
    geneva_k8s_service_account: str,
    k8s_config_method: K8sConfigMethod,
    k8s_namespace: str,
    region: str,
    head_node_selector: dict,
    worker_node_selector: dict,
    k8s_cluster_name: str,
    slug: str,
) -> str:
    """
    Define a standard Ray cluster for e2e tests.

    Returns the cluster name for use with conn.context(cluster=name, manifest=name).
    """
    conn, _, _ = oxford_pets_table
    cluster_name = "e2e-standard-cluster"

    _LOG.info(f"Defining standard cluster '{cluster_name}'")

    # Define cluster configuration
    # Pin to ray 2.44.0 to match live cluster environments
    # TODO: When live clusters upgrade, update this version
    head_group = HeadGroupConfig(
        service_account=geneva_k8s_service_account,
        num_cpus=1,
        memory="3Gi",
        image="rayproject/ray:2.44.0-py310",
        node_selector=head_node_selector,
        labels={},
        tolerations=[],
    )

    worker_group = WorkerGroupConfig(
        service_account=geneva_k8s_service_account,
        num_cpus=2,
        memory="4Gi",
        image="rayproject/ray:2.44.0-py310",
        node_selector=worker_node_selector,
        labels={},
        tolerations=[],
    )

    kuberay_config = KubeRayConfig(
        namespace=k8s_namespace,
        head_group=head_group,
        worker_groups=[worker_group],
        config_method=k8s_config_method,
        use_portforwarding=True,
        aws_region=region if k8s_config_method == K8sConfigMethod.EKS_AUTH else None,
        aws_role_name="geneva-client-role" if k8s_config_method == K8sConfigMethod.EKS_AUTH else None,
    )

    cluster = GenevaCluster(
        cluster_type=GenevaClusterType.KUBE_RAY,
        name=cluster_name,
        kuberay=kuberay_config,
    )

    conn.define_cluster(cluster_name, cluster)
    _LOG.info(f"Cluster '{cluster_name}' defined")

    return cluster_name


@pytest.fixture
def gpu_cluster(
    oxford_pets_table: tuple,
    geneva_k8s_service_account: str,
    k8s_config_method: K8sConfigMethod,
    k8s_namespace: str,
    csp: str,
    region: str,
    head_node_selector: dict,
    k8s_cluster_name: str,
    slug: str,
) -> str:
    """
    Define a GPU Ray cluster for caption and embedding generation.

    Returns the cluster name for use with conn.context(cluster=name, manifest=name).
    """
    conn, _, _ = oxford_pets_table
    cluster_name = "e2e-gpu-cluster"

    _LOG.info(f"Defining GPU cluster '{cluster_name}'")

    # GPU worker node selector
    gpu_worker_node_selector = (
        {"geneva.lancedb.com/ray-worker-gpu": "true"}
        if csp == "aws"
        else {"_PLACEHOLDER": "true"}
    )

    # Define cluster configuration
    # Pin to ray 2.44.0 to match live cluster environments
    # TODO: When live clusters upgrade, update this version
    head_group = HeadGroupConfig(
        service_account=geneva_k8s_service_account,
        num_cpus=1,
        memory="3Gi",
        image="rayproject/ray:2.44.0-py310",
        node_selector=head_node_selector,
        labels={},
        tolerations=[],
    )

    worker_group = WorkerGroupConfig(
        service_account=geneva_k8s_service_account,
        num_cpus=4,
        memory="16Gi",
        image="rayproject/ray:2.44.0-py310",
        node_selector=gpu_worker_node_selector,
        labels={},
        tolerations=[],
        num_gpus=1,
    )

    kuberay_config = KubeRayConfig(
        namespace=k8s_namespace,
        head_group=head_group,
        worker_groups=[worker_group],
        config_method=k8s_config_method,
        use_portforwarding=True,
        aws_region=region if k8s_config_method == K8sConfigMethod.EKS_AUTH else None,
        aws_role_name="geneva-client-role" if k8s_config_method == K8sConfigMethod.EKS_AUTH else None,
    )

    cluster = GenevaCluster(
        cluster_type=GenevaClusterType.KUBE_RAY,
        name=cluster_name,
        kuberay=kuberay_config,
    )

    conn.define_cluster(cluster_name, cluster)
    _LOG.info(f"Cluster '{cluster_name}' defined")

    return cluster_name



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors
from collections.abc import Generator
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock, patch

import pytest
import ray
import yaml

try:
    from geneva.runners.ray.raycluster import (
        ClusterStatus,
        RayCluster,
        _HeadGroupSpec,
        _WorkerGroupSpec,
    )

except ImportError:
    pytest.skip("failed to import geneva.runners.ray", allow_module_level=True)

if TYPE_CHECKING:
    from geneva.runners.ray.kuberay import KuberaySummary

# Disable real k8s client init in attrs post_init for serialization tests
RayCluster.__attrs_post_init__ = lambda self: None


@pytest.fixture(autouse=True)
def ensure_ray_shutdown() -> Generator[None, None, None]:
    """Ensure Ray is shut down before and after each test"""
    if ray.is_initialized():
        ray.shutdown()
    yield
    if ray.is_initialized():
        ray.shutdown()


class DummyCM:
    def __init__(self, data: dict[str, str]) -> None:
        self.data = data


def test_env_vars_in_head_and_worker() -> None:
    """Test that env_vars can be set on both head and worker groups and
    appear in pod definition"""
    head = _HeadGroupSpec(
        env_vars={
            "GCS_REQUEST_CONNECTION_TIMEOUT_SECS": "300",
            "GCS_REQUEST_TIMEOUT_SECS": "600",
            "CUSTOM_VAR": "value",
        }
    )
    worker = _WorkerGroupSpec(
        env_vars={
            "GCS_REQUEST_CONNECTION_TIMEOUT_SECS": "300",
            "GCS_REQUEST_TIMEOUT_SECS": "600",
            "WORKER_VAR": "worker_value",
        }
    )

    cluster = RayCluster(
        name="test-env-vars", namespace="test", head_group=head, worker_groups=[worker]
    )

    # Check head group container env vars in the generated definition
    head_spec = cluster.definition["spec"]["headGroupSpec"]
    head_container = head_spec["template"]["spec"]["containers"][0]
    assert "env" in head_container
    head_env = {env["name"]: env["value"] for env in head_container["env"]}
    assert head_env["GCS_REQUEST_CONNECTION_TIMEOUT_SECS"] == "300"
    assert head_env["GCS_REQUEST_TIMEOUT_SECS"] == "600"
    assert head_env["CUSTOM_VAR"] == "value"
    assert len(head_env) == 3

    # Check worker group container env vars in the generated definition
    worker_spec = cluster.definition["spec"]["workerGroupSpecs"][0]
    worker_container = worker_spec["template"]["spec"]["containers"][0]
    assert "env" in worker_container
    worker_env = {env["name"]: env["value"] for env in worker_container["env"]}
    # Worker has 2 default env vars (RAY_memory_usage_threshold,
    # RAY_memory_monitor_refresh_ms) plus the 3 custom ones
    assert worker_env["GCS_REQUEST_CONNECTION_TIMEOUT_SECS"] == "300"
    assert worker_env["GCS_REQUEST_TIMEOUT_SECS"] == "600"
    assert worker_env["WORKER_VAR"] == "worker_value"
    assert worker_env["RAY_memory_usage_threshold"] == "0.9"
    assert worker_env["RAY_memory_monitor_refresh_ms"] == "0"
    assert len(worker_env) == 5


def test_env_vars_empty() -> None:
    """Test that empty env_vars dict produces empty env list"""
    head = _HeadGroupSpec(env_vars={})
    worker = _WorkerGroupSpec(env_vars={})

    cluster = RayCluster(
        name="test-empty-env", namespace="test", head_group=head, worker_groups=[worker]
    )

    # Check head group has empty env list
    head_spec = cluster.definition["spec"]["headGroupSpec"]
    head_container = head_spec["template"]["spec"]["containers"][0]
    assert head_container["env"] == []

    # Check worker group has only default env vars
    worker_spec = cluster.definition["spec"]["workerGroupSpecs"][0]
    worker_container = worker_spec["template"]["spec"]["containers"][0]
    worker_env = {env["name"]: env["value"] for env in worker_container["env"]}
    assert worker_env["RAY_memory_usage_threshold"] == "0.9"
    assert worker_env["RAY_memory_monitor_refresh_ms"] == "0"
    assert len(worker_env) == 2


def test_env_vars_serialization_to_config_map() -> None:
    """Test that env_vars are preserved in ConfigMap serialization"""
    head = _HeadGroupSpec(
        image="img",
        num_cpus=1,
        memory="4Gi",
        env_vars={"HEAD_VAR": "head_value"},
    )
    worker = _WorkerGroupSpec(
        image="img",
        num_cpus=2,
        memory="8Gi",
        env_vars={"WORKER_VAR": "worker_value"},
    )

    _ = RayCluster(
        name="test-env-serialization",
        namespace="ns",
        head_group=head,
        worker_groups=[worker],
    )


def test_cluster_status_get_status_happy_path() -> None:
    head = _HeadGroupSpec(image="test-img", num_cpus=1, memory="4Gi")
    worker = _WorkerGroupSpec(image="test-img", num_cpus=2, memory="8Gi")

    mock_ray_cluster = RayCluster(
        name="test-cluster",
        namespace="test-namespace",
        head_group=head,
        worker_groups=[worker],
    )

    mock_clients = Mock()
    mock_ray_cluster.clients = mock_clients

    mock_summary: KuberaySummary = {
        "phase": "Running",
        "phase_idx": 3,
        "pending": 0,
        "running": 3,
        "workers_ready": 2,
        "workers_ready_gpu": 1,
        "workers_ready_cpu": 1,
        "pods_gpu_running": 1,
        "pods_gpu_pending": 0,
        "pods_cpu_running": 2,
        "pods_cpu_pending": 0,
        "waits_top3": [],
        "pulling": 0,
        "total_pods": 3,
        "kr_state": "running",
        "kr_desired_workers": 2,
        "kr_available_workers": 2,
        "kr_scaling": "steady",
        "kr_last_condition": ("Ready", "True"),
        "nodes_gpu_ready": 1,
        "nodes_gpu_notready": 0,
        "nodes_cpu_ready": 1,
        "nodes_cpu_notready": 0,
    }

    cluster_status = ClusterStatus()
    cluster_status.namespace = "test-namespace"
    cluster_status.cluster_name = "test-cluster"
    cluster_status.ray_cluster = mock_ray_cluster

    # Mock the summarize_kuberay_status function
    with patch(
        "geneva.runners.ray.raycluster.summarize_kuberay_status"
    ) as mock_summarize:
        mock_summarize.return_value = mock_summary

        # Mock tqdm to avoid actual progress bar creation
        with patch("geneva.runners.ray.raycluster.tqdm") as mock_tqdm:
            mock_pbar = Mock()
            mock_tqdm.return_value = mock_pbar

            # Call get_status()
            cluster_status.get_status()

            # Verify summarize_kuberay_status was called with correct parameters
            mock_summarize.assert_called_once_with(
                mock_clients, "test-namespace", "test-cluster"
            )

            # Verify progress bars were created
            assert cluster_status.pbar_k8s is not None
            assert cluster_status.pbar_kuberay is not None

            # Verify progress bar refresh was called for both bars
            assert mock_pbar.refresh.call_count == 2  # Called for both k8s and kuberay

            # Verify that progress bar descriptions were set
            # The desc attribute should have been set
            assert hasattr(mock_pbar, "desc")

            # Verify that tqdm was called twice (once for k8s, once for kuberay)
            assert mock_tqdm.call_count == 2


def test_env_vars_serialization_to_config_map_continued() -> None:
    """Test that env_vars are preserved in ConfigMap serialization - continuation"""
    head = _HeadGroupSpec(
        image="img",
        num_cpus=1,
        memory="4Gi",
        env_vars={"HEAD_VAR": "head_value"},
    )
    worker = _WorkerGroupSpec(
        image="img",
        num_cpus=2,
        memory="8Gi",
        env_vars={"WORKER_VAR": "worker_value"},
    )

    cluster = RayCluster(
        name="test-env-serialization",
        namespace="ns",
        head_group=head,
        worker_groups=[worker],
    )

    data = cluster.to_config_map()
    h = yaml.safe_load(data["head_group"])
    assert h["env_vars"] == {"HEAD_VAR": "head_value"}

    ws = yaml.safe_load(data["worker_groups"])
    assert ws[0]["env_vars"] == {"WORKER_VAR": "worker_value"}


def test_env_vars_deserialization_from_config_map(monkeypatch) -> None:
    """Test that env_vars are properly loaded from ConfigMap"""
    dummy_data = {
        "name": "env-vars-cluster",
        "head_group": yaml.dump(
            {
                "image": "ray:latest",
                "num_cpus": 2,
                "memory": "4Gi",
                "env_vars": {"HEAD_ENV": "head_val"},
            }
        ),
        "worker_groups": yaml.dump(
            [
                {
                    "image": "ray:latest",
                    "num_cpus": 4,
                    "memory": "8Gi",
                    "env_vars": {"WORKER_ENV": "worker_val"},
                }
            ]
        ),
    }
    cm = DummyCM(dummy_data)

    monkeypatch.setattr(
        "geneva.runners.kuberay.client.build_api_client",
        lambda *args: None,
    )

    class DummyCore:
        def read_namespaced_config_map(self, name: str, namespace: str) -> DummyCM:
            return cm

    monkeypatch.setattr(
        "geneva.runners.ray.raycluster.kubernetes.client.CoreV1Api",
        lambda api_client=None: DummyCore(),
    )

    cluster = RayCluster.from_config_map(
        "my-namespace", "my-k8s-cluster", "cm-name", "env-vars-cluster"
    )

    assert cluster.head_group.env_vars == {"HEAD_ENV": "head_val"}
    assert cluster.worker_groups[0].env_vars == {"WORKER_ENV": "worker_val"}


def test_extra_env_passed_to_init_ray(monkeypatch) -> None:
    """Test that extra_env parameter is passed through ray_cluster to init_ray

    This test verifies that extra_env variables are correctly passed through the
    ray_cluster context manager to init_ray, where they become part of Ray's
    runtime environment. These variables will be available in Ray worker processes
    when executing tasks/actors on both head and worker nodes.
    """

    import geneva.runners.ray._mgr as ray_mgr_mod

    ray_init_called = {}

    # Mock ray.init to capture the runtime_env
    def mock_ray_init(*args: Any, **kwargs: Any) -> None:
        ray_init_called["args"] = args
        ray_init_called["kwargs"] = kwargs

    monkeypatch.setattr("ray.init", mock_ray_init)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    # Test with local ray cluster to avoid kubernetes dependencies
    extra_env_vars = {
        "GCS_REQUEST_CONNECTION_TIMEOUT_SECS": "300",
        "GCS_REQUEST_TIMEOUT_SECS": "600",
        "CUSTOM_ENV_VAR": "custom_value",
    }

    with ray_mgr_mod.ray_cluster(local=True, extra_env=extra_env_vars):
        pass

    # Verify ray.init was called with correct runtime_env
    assert "kwargs" in ray_init_called
    runtime_env = ray_init_called["kwargs"]["runtime_env"]

    # Check that extra_env variables are in the runtime_env
    assert "env_vars" in runtime_env
    env_vars = runtime_env["env_vars"]

    # Verify our custom env vars are present
    assert env_vars["GCS_REQUEST_CONNECTION_TIMEOUT_SECS"] == "300"
    assert env_vars["GCS_REQUEST_TIMEOUT_SECS"] == "600"
    assert env_vars["CUSTOM_ENV_VAR"] == "custom_value"

    # GENEVA_ZIPS should also be present (added by init_ray)
    assert "GENEVA_ZIPS" in env_vars


def test_env_vars_vs_extra_env(monkeypatch) -> None:
    """Test interaction between env_vars (in head spec) and extra_env

    This test demonstrates that:
    - env_vars in head/worker specs set Kubernetes container environment variables
    - extra_env in ray_cluster sets Ray runtime_env environment variables
    - When the same variable is set in both places, extra_env takes
      precedence in Ray tasks
    - extra_env is what gets passed to init_ray's runtime_env, not env_vars
      from specs
    """

    import geneva.runners.ray._mgr as ray_mgr_mod

    ray_init_called = {}

    # Mock ray.init to capture the runtime_env
    def mock_ray_init(*args: Any, **kwargs: Any) -> None:
        ray_init_called["args"] = args
        ray_init_called["kwargs"] = kwargs

    monkeypatch.setattr("ray.init", mock_ray_init)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    # Create a RayCluster with env_vars in the head spec
    # These would normally go into the Kubernetes pod definition
    head = _HeadGroupSpec(
        env_vars={
            "SHARED_VAR": "from_head_spec",
            "HEAD_ONLY_VAR": "head_value",
        }
    )
    # Note: cluster object is not used, but demonstrates that env_vars
    # in head spec don't automatically flow to runtime_env
    _ = RayCluster(
        name="test-env-conflict",
        namespace="test",
        head_group=head,
        worker_groups=[],
    )

    # Now use ray_cluster with extra_env
    # These go into Ray's runtime_env
    extra_env_vars = {
        "SHARED_VAR": "from_extra_env",  # Overlaps with head spec
        "EXTRA_ONLY_VAR": "extra_value",  # Only in extra_env
    }

    # Note: We use local=True to avoid actually creating a k8s cluster
    # In a real scenario with a k8s cluster, both env_vars (in containers)
    # and extra_env (in Ray runtime_env) would be present
    with ray_mgr_mod.ray_cluster(local=True, extra_env=extra_env_vars):
        pass

    # Verify what ended up in the runtime_env
    assert "kwargs" in ray_init_called
    runtime_env = ray_init_called["kwargs"]["runtime_env"]
    assert "env_vars" in runtime_env
    env_vars = runtime_env["env_vars"]

    # Key insight: Only extra_env makes it to the runtime_env
    # The env_vars from head spec are NOT in the runtime_env
    # (they would only be in the Kubernetes pod definition)
    assert env_vars["SHARED_VAR"] == "from_extra_env"  # extra_env wins for Ray tasks
    assert env_vars["EXTRA_ONLY_VAR"] == "extra_value"
    assert "HEAD_ONLY_VAR" not in env_vars  # Not in runtime_env
    assert "GENEVA_ZIPS" in env_vars  # Always present


@pytest.mark.ray
def test_extra_env_available_in_ray_workers() -> None:
    """Test that extra_env variables are actually available in Ray worker processes

    This integration test starts a real local Ray cluster and verifies that
    environment variables passed via extra_env are accessible in Ray remote
    functions running on worker processes.
    """
    import os

    import ray

    import geneva.runners.ray._mgr as ray_mgr_mod

    # Define a remote function that checks environment variables
    @ray.remote
    def check_env_vars() -> dict[str, str | None]:
        """Remote function that runs in a Ray worker and checks env vars"""
        return {
            "GCS_REQUEST_CONNECTION_TIMEOUT_SECS": os.environ.get(
                "GCS_REQUEST_CONNECTION_TIMEOUT_SECS"
            ),
            "GCS_REQUEST_TIMEOUT_SECS": os.environ.get("GCS_REQUEST_TIMEOUT_SECS"),
            "CUSTOM_ENV_VAR": os.environ.get("CUSTOM_ENV_VAR"),
        }

    extra_env_vars = {
        "GCS_REQUEST_CONNECTION_TIMEOUT_SECS": "300",
        "GCS_REQUEST_TIMEOUT_SECS": "600",
        "CUSTOM_ENV_VAR": "custom_value",
    }

    # Use local Ray cluster to test
    with ray_mgr_mod.ray_cluster(local=True, extra_env=extra_env_vars):
        # Execute the remote function and get the result
        result = ray.get(check_env_vars.remote())

        # Verify all env vars are accessible in the worker
        assert result["GCS_REQUEST_CONNECTION_TIMEOUT_SECS"] == "300"
        assert result["GCS_REQUEST_TIMEOUT_SECS"] == "600"
        assert result["CUSTOM_ENV_VAR"] == "custom_value"


def test_ray_init_kwargs_passed_to_init_ray(monkeypatch) -> None:
    """Test that ray_init_kwargs parameter is passed through to ray.init()

    This test verifies that arbitrary kwargs passed via ray_init_kwargs are
    correctly threaded through ray_cluster -> init_ray -> ray.init().
    """

    import geneva.runners.ray._mgr as ray_mgr_mod

    ray_init_called = {}

    # Mock ray.init to capture all kwargs
    def mock_ray_init(*args: Any, **kwargs: Any) -> None:
        ray_init_called["args"] = args
        ray_init_called["kwargs"] = kwargs

    monkeypatch.setattr("ray.init", mock_ray_init)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    # Test with custom runtime_env via ray_init_kwargs
    custom_runtime_env = {
        "conda": {
            "channels": ["conda-forge"],
            "dependencies": ["python=3.10", "ffmpeg"],
        },
        "config": {"eager_install": True},
    }

    ray_init_kwargs = {
        "runtime_env": custom_runtime_env,
        "namespace": "test-namespace",
        "ignore_reinit_error": True,
    }

    with ray_mgr_mod.ray_cluster(local=True, ray_init_kwargs=ray_init_kwargs):
        pass

    # Verify ray.init was called with our custom kwargs
    assert "kwargs" in ray_init_called
    kwargs = ray_init_called["kwargs"]

    # Check that ray_init_kwargs are present
    assert kwargs["namespace"] == "test-namespace"
    assert kwargs["ignore_reinit_error"] is True

    # Check that runtime_env was merged correctly
    runtime_env = kwargs["runtime_env"]
    assert "conda" in runtime_env
    assert runtime_env["conda"] == custom_runtime_env["conda"]
    assert runtime_env["config"] == custom_runtime_env["config"]

    # GENEVA_ZIPS should still be present in env_vars
    assert "env_vars" in runtime_env
    assert "GENEVA_ZIPS" in runtime_env["env_vars"]


def test_ray_init_kwargs_runtime_env_merge(monkeypatch) -> None:
    """Test that ray_init_kwargs runtime_env is properly merged with default runtime_env

    This test verifies that when ray_init_kwargs contains a runtime_env,
    it's properly merged with the default Geneva runtime_env, preserving
    both user settings and Geneva internals.
    """

    import geneva.runners.ray._mgr as ray_mgr_mod

    ray_init_called = {}

    def mock_ray_init(*args: Any, **kwargs: Any) -> None:
        ray_init_called["kwargs"] = kwargs

    monkeypatch.setattr("ray.init", mock_ray_init)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    # Use both extra_env and ray_init_kwargs with runtime_env
    extra_env_vars = {
        "EXTRA_VAR": "extra_value",
    }

    ray_init_kwargs = {
        "runtime_env": {
            "env_vars": {
                "CUSTOM_VAR": "custom_value",
            },
            "pip": ["numpy==1.24.0"],
        }
    }

    with ray_mgr_mod.ray_cluster(
        local=True, extra_env=extra_env_vars, ray_init_kwargs=ray_init_kwargs
    ):
        pass

    runtime_env = ray_init_called["kwargs"]["runtime_env"]

    # Check that all env_vars are merged
    assert "GENEVA_ZIPS" in runtime_env["env_vars"]  # Default from Geneva
    assert runtime_env["env_vars"]["EXTRA_VAR"] == "extra_value"  # From extra_env
    assert (
        runtime_env["env_vars"]["CUSTOM_VAR"] == "custom_value"
    )  # From ray_init_kwargs

    # Check that pip from ray_init_kwargs is preserved
    assert runtime_env["pip"] == ["numpy==1.24.0"]


def test_raycluster_ray_init_kwargs_attribute() -> None:
    """Test that RayCluster properly stores ray_init_kwargs"""
    head = _HeadGroupSpec(image="test-img", num_cpus=1, memory="4Gi")
    worker = _WorkerGroupSpec(image="test-img", num_cpus=2, memory="8Gi")

    ray_init_kwargs = {
        "runtime_env": {
            "conda": {"dependencies": ["python=3.10"]},
        },
        "namespace": "my-namespace",
    }

    cluster = RayCluster(
        name="test-cluster",
        namespace="test-ns",
        head_group=head,
        worker_groups=[worker],
        ray_init_kwargs=ray_init_kwargs,
    )

    assert cluster.ray_init_kwargs == ray_init_kwargs
    assert cluster.ray_init_kwargs["namespace"] == "my-namespace"


def test_geneva_cluster_ray_init_kwargs() -> None:
    """Test that GenevaCluster properly stores and converts ray_init_kwargs"""
    from geneva.cluster import GenevaClusterType, K8sConfigMethod
    from geneva.cluster.mgr import GenevaCluster, HeadGroupConfig, KubeRayConfig

    ray_init_kwargs = {
        "runtime_env": {
            "pip": ["numpy"],
        },
        "namespace": "test-ns",
    }

    head_config = HeadGroupConfig(
        service_account="test-sa",
        num_cpus=2,
        memory="4Gi",
        image="rayproject/ray:latest",
        node_selector={},
        labels={},
        tolerations=[],
    )

    kuberay_config = KubeRayConfig(
        namespace="geneva",
        head_group=head_config,
        worker_groups=[],
        config_method=K8sConfigMethod.LOCAL,
        ray_init_kwargs=ray_init_kwargs,
    )

    cluster = GenevaCluster(
        cluster_type=GenevaClusterType.KUBE_RAY,
        name="test-cluster",
        kuberay=kuberay_config,
    )

    # Verify ray_init_kwargs are stored
    assert cluster.kuberay.ray_init_kwargs == ray_init_kwargs

    # Verify conversion to RayCluster preserves ray_init_kwargs
    ray_cluster = cluster.to_ray_cluster()
    assert ray_cluster.ray_init_kwargs == ray_init_kwargs


def test_conda_pip_conflict(monkeypatch) -> None:
    """Test handling when both conda and pip are specified

    According to Ray documentation, conda and pip cannot be specified simultaneously.
    This test verifies the behavior when both are present.
    """
    import geneva.runners.ray._mgr as ray_mgr_mod

    ray_init_called = {}

    def mock_ray_init(*args: Any, **kwargs: Any) -> None:
        ray_init_called["kwargs"] = kwargs

    monkeypatch.setattr("ray.init", mock_ray_init)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    # Geneva sets pip via parameter, user sets conda via ray_init_kwargs
    pip_packages = ["numpy==1.24.0"]
    ray_init_kwargs = {
        "runtime_env": {
            "conda": {
                "channels": ["conda-forge"],
                "dependencies": ["python=3.10", "ffmpeg"],
            }
        }
    }

    with ray_mgr_mod.ray_cluster(
        local=True, pip=pip_packages, ray_init_kwargs=ray_init_kwargs
    ):
        pass

    runtime_env = ray_init_called["kwargs"]["runtime_env"]

    # Both pip and conda will be present - this is expected behavior
    # Ray will validate and raise an error when trying to use this runtime_env
    assert "pip" in runtime_env
    assert "conda" in runtime_env
    # Note: In real usage, Ray would reject this configuration


def test_conda_overrides_pip_in_ray_init_kwargs(monkeypatch) -> None:
    """Test that conda in ray_init_kwargs can override pip at the same level

    If user specifies both pip via parameter AND conda in ray_init_kwargs,
    they can also override by putting pip in ray_init_kwargs.
    """
    import geneva.runners.ray._mgr as ray_mgr_mod

    ray_init_called = {}

    def mock_ray_init(*args: Any, **kwargs: Any) -> None:
        ray_init_called["kwargs"] = kwargs

    monkeypatch.setattr("ray.init", mock_ray_init)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    # User can override pip by passing pip: None in ray_init_kwargs
    pip_packages = ["numpy==1.24.0"]
    ray_init_kwargs = {
        "runtime_env": {
            "pip": None,  # Explicitly clear pip
            "conda": {
                "channels": ["conda-forge"],
                "dependencies": ["python=3.10", "ffmpeg"],
            },
        }
    }

    with ray_mgr_mod.ray_cluster(
        local=True, pip=pip_packages, ray_init_kwargs=ray_init_kwargs
    ):
        pass

    runtime_env = ray_init_called["kwargs"]["runtime_env"]

    # pip should be None, conda should be present
    assert runtime_env.get("pip") is None
    assert "conda" in runtime_env


def test_container_runtime_env(monkeypatch) -> None:
    """Test that container in runtime_env is passed through correctly

    According to Ray docs, container works alone or only with config/env_vars.
    """
    import geneva.runners.ray._mgr as ray_mgr_mod

    ray_init_called = {}

    def mock_ray_init(*args: Any, **kwargs: Any) -> None:
        ray_init_called["kwargs"] = kwargs

    monkeypatch.setattr("ray.init", mock_ray_init)
    monkeypatch.setattr("ray.shutdown", lambda: None)
    monkeypatch.setattr("ray.is_initialized", lambda: False)

    ray_init_kwargs = {
        "runtime_env": {
            "container": {
                "image": "anyscale/ray:latest",
                "run_options": ["--cap-add=SYS_PTRACE"],
            },
        }
    }

    with ray_mgr_mod.ray_cluster(local=True, ray_init_kwargs=ray_init_kwargs):
        pass

    runtime_env = ray_init_called["kwargs"]["runtime_env"]

    # Container should be present
    assert "container" in runtime_env
    assert runtime_env["container"]["image"] == "anyscale/ray:latest"
    # env_vars (GENEVA_ZIPS) should still be there
    assert "env_vars" in runtime_env
    assert "GENEVA_ZIPS" in runtime_env["env_vars"]



# SPDX-License-Identifier: PROPRIETARY
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# base class for configuration syntax sugar

import abc


class KV(abc.ABC):
    @abc.abstractmethod
    def __getitem__(self, key: list[str]) -> str:
        pass

    @abc.abstractmethod
    def __contains__(self, key: list[str]) -> bool:
        """Populate the configuration from a dictionary"""
